"""Command-line interface options for the readme-ai package."""

from __future__ import annotations

import click

from readmeai import __version__
from readmeai.config.constants import (
    BadgeStyleOptions,
    HeaderStyleOptions,
    ImageOptions,
    LLMService,
)


def prompt_for_image(
    ctx: click.Context | None = None,
    param: click.Parameter | None = None,
    value: str | None = None,
) -> str:
    """Prompt the user for a custom image URL."""
    if value is None:
        return ImageOptions.BLUE.value
    if value == ImageOptions.CUSTOM.name:
        return click.prompt("Provide an image file path or URL")
    elif value == ImageOptions.LLM.name:
        return ImageOptions.LLM.value
    elif value in ImageOptions.__members__:
        return ImageOptions[value].value
    else:
        raise click.BadParameter(f"Invalid image provided: {value}")


def version_callback(
    ctx: click.Context | None = None,
    param: click.Parameter | None = None,
    value: str | None = None,
) -> None:
    """Prints the version of readme-ai."""
    if not value or (ctx and ctx.resilient_parsing):
        return
    click.echo(f"readmeai version {__version__}")
    if ctx is not None:
        ctx.exit()


align = click.option(
    "-a",
    "--align",
    type=click.Choice(["center", "left", "right"], case_sensitive=False),
    default="center",
    help="align for the README.md file header sections.",
)

api = click.option(
    "--api",
    type=click.Choice(
        [opt.value for opt in LLMService],
        case_sensitive=False,
    ),
    default=LLMService.OFFLINE,
    help="""LLM API service to use for README.md file generation. Current support for:\n
    - ANTHROPIC    # Anthropic: claude-3-5-sonnet \n
    - GEMINI       # Google Gemini: gemini-pro \n
    - OFFLINE      # Offline Mode: run without a LLM API \n
    - OLLAMA       # Ollama: llama3, mistral, etc. \n
    - OPENAI       # OpenAI: gpt-3.5-turbo \n
    """,
)

badge_color = click.option(
    "-bc",
    "--badge-color",
    type=str,
    default="0080ff",
    help="Change color of shields.io badges. Provide color name or hex code.",
)

badge_style = click.option(
    "-bs",
    "--badge-style",
    type=click.Choice(
        [opt.value for opt in BadgeStyleOptions],
        case_sensitive=False,
    ),
    default=BadgeStyleOptions.DEFAULT.value,
    help="""\
        Style for shields.io badges. Current support for:\n
        - default \n
        - flat \n
        - flat-square \n
        - for-the-badge \n
        - plastic \n
        - skills \n
        - skills-light \n
        - social \n
        """,
)

base_url = click.option(
    "--base-url",
    type=str,
    default="https://api.openai.com/v1/chat/completions",
    help="Base URL for the LLM API service.",
)

context_window = click.option(
    "-cw",
    "--context-window",
    default=3900,
    type=int,
    help="Maximum number of tokens to use for the model's context window.",
)

emojis = click.option(
    "-e",
    "--emojis",
    is_flag=True,
    default=False,
    help="Adds an emoji prefix to each header of the README.md file. For example, the header ## 'Overview' would change to '## 📍 Overview'.",
)

header_style = click.option(
    "-hs",
    "--header-style",
    type=click.Choice(
        [opt.name for opt in HeaderStyleOptions],
        case_sensitive=False,
    ),
    default="classic",
    help="Header template styles.",
)

toc_style = click.option(
    "-ts",
    "--toc-style",
    type=click.Choice(
        ["bullet", "fold", "links", "number", "roman"],
        case_sensitive=False,
    ),
    default="bullet",
    help="Table of Contents template styles.",
)

image = click.option(
    "-i",
    "--image",
    type=click.Choice(
        [opt.name for opt in ImageOptions],
        case_sensitive=False,
    ),
    default=ImageOptions.BLUE.name,
    callback=prompt_for_image,
    show_choices=True,
    help="""\
        Project logo image displayed in the README file header. The following options are currently supported:\n
        - CUSTOM (provide local image path or URL) \n
        - LLM (generate project logo using LLM API) \n
        - BLACK \n
        - BLUE \n
        - CLOUD \n
        - GRADIENT \n
        - GREY \n
        - PURPLE \n
        """,
)

model = click.option(
    "-m",
    "--model",
    default="gpt-3.5-turbo",
    help="LLM API service to use for README file text generation.",
)

output = click.option(
    "-o",
    "--output",
    default="readme-ai.md",
    help="Output file name for your README file.",
)

rate_limit = click.option(
    "-rl",
    "--rate-limit",
    default=10,
    type=click.IntRange(1, 25, clamp=True),
    help="Number of requests per minute for the LLM API service.",
)

repository = click.option(
    "-r",
    "--repository",
    required=True,
    help="Provide a remote repository URL (GitHub, GitLab, BitBucket), or local path to your project.",
)

temperature = click.option(
    "-t",
    "--temperature",
    default=0.1,
    type=click.FloatRange(min_open=0.0, max=2.0, clamp=True),
    help="Increasing temperature yields more randomness in text generation.",
)

top_p = click.option(
    "--top-p",
    default=0.9,
    type=click.FloatRange(0.0, 1.0, clamp=True),
    help="Top-p sampling probability for the model's generation.",
)

tree_depth = click.option(
    "-td",
    "--tree-depth",
    default=2,
    type=int,
    help="Maximum depth of the directory tree generated for the README file.",
)



"""Parser for Docker (Dockerfile, docker-compose.yaml) configuration files."""

import re
from typing import Any

import yaml

from readmeai.parsers.base import BaseFileParser


class DockerfileParser(BaseFileParser):
    """
    Parser for Dockerfile dependency files.
    """

    def __init__(self) -> None:
        super().__init__()

    def parse(self, content: str) -> list[str]:
        """Extracts package names from a Dockerfile."""
        try:
            dependencies = []
            lines = content.split("\n")
            for line in lines:
                if match := re.search(
                    r"^FROM\s+(?:--platform=[^\s]+\s+)?([^\s:]+):?([^\s]*)",
                    line,
                ):
                    base_image, version = match.groups()
                    if not version:
                        version = "latest"
                    dependencies.append((base_image, version))

            return [f"{dep[0]}: {dep[1]}" for dep in dependencies]

        except re.error as e:
            return self.handle_parsing_error(e)


class DockerComposeParser(BaseFileParser):
    """
    Parser for docker-compose.yaml files.
    """

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()
        self.compose_data: dict[str, Any] = {}

    def parse(self, content: str) -> list[str]:
        """Parses docker-compose YAML content and returns dependency list."""
        try:
            self.compose_data = yaml.safe_load(content)
            return self.get_services()
        except yaml.YAMLError as e:
            return self.handle_parsing_error(e)

    def get_services(self) -> list[str]:
        """Get a list of all service names from the docker-compose.yaml file."""
        if self.compose_data is None:
            return []
        return list(self.compose_data.get("services", {}).keys())

    def get_service_info(self, service_name: str) -> dict[str, Any]:
        """Get detailed information for a specific service."""
        return self.compose_data.get("services", {}).get(service_name, {})

    def get_service_environment(self, service_name: str) -> list[str]:
        """Get the environment variables for a specific service."""
        service_info = self.get_service_info(service_name)
        return service_info.get("environment", [])

    def get_service_ports(self, service_name: str) -> list[str]:
        """Get the ports for a specific service."""
        service_info = self.get_service_info(service_name)
        return service_info.get("ports", [])

    def get_service_command(self, service_name: str) -> str:
        """Get the command for a specific service."""
        service_info = self.get_service_info(service_name)
        return service_info.get("command", "")

    def get_service_networks(self, service_name: str) -> list[str]:
        """Get the networks for a specific service."""
        service_info = self.get_service_info(service_name)
        return service_info.get("networks", [])

    def get_service_image(self, service_name: str) -> str:
        """Get the image used by a specific service."""
        service_info = self.get_service_info(service_name)
        return service_info.get("image", "")

    def get_all_service_details(self) -> list[dict[str, dict[str, Any]]]:
        """Get all metadata for all services."""
        return [
            {
                service_name: {
                    "image": self.get_service_image(service_name),
                    "environment": self.get_service_environment(service_name),
                    "ports": self.get_service_ports(service_name),
                    "command": self.get_service_command(service_name),
                    "networks": self.get_service_networks(service_name),
                    "details": self.get_service_info(service_name),
                }
                for service_name in self.get_services()
            }
        ]



"""Anthropic API service implementation."""

import os
from typing import Any, Optional

from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from readmeai.config.settings import ConfigLoader
from readmeai.ingestion.models import RepositoryContext
from readmeai.models.base import BaseModelHandler
from readmeai.models.tokens import token_handler

try:
    import anthropic

    ANTHROPIC_AVAILABLE = True
    ANTHROPIC_EXCEPTIONS = (
        anthropic.APIError,
        anthropic.APIConnectionError,
        anthropic.RateLimitError,
    )
except ImportError:
    anthropic = None
    ANTHROPIC_AVAILABLE = False
    ANTHROPIC_EXCEPTIONS = tuple()


class AnthropicHandler(BaseModelHandler):
    """
    Anthropic Claude LLM API service implementation.
    """

    def __init__(
        self, config_loader: ConfigLoader, context: RepositoryContext
    ) -> None:
        super().__init__(config_loader, context)
        self.client: Optional[Any] = None
        self.model: str = "claude-3-opus-20240229"
        if ANTHROPIC_AVAILABLE:
            self._model_settings()
        else:
            self._logger.warning(
                "Anthropic library is not available. Some features will be disabled."
            )

    def _model_settings(self):
        if not ANTHROPIC_AVAILABLE:
            self._logger.error(
                "Attempted to configure Anthropic model without the required library."
            )
            return

        api_key = os.getenv("ANTHROPIC_API_KEY")
        if not api_key:
            self._logger.error(
                "ANTHROPIC_API_KEY environment variable is not set."
            )
            return

        self.client = anthropic.AsyncAnthropic(api_key=api_key)

    async def _build_payload(self, prompt: str, tokens: int) -> dict[str, Any]:
        """Build payload for POST request to the Anthropic API."""
        if not ANTHROPIC_AVAILABLE:
            raise RuntimeError("Anthropic library is not available.")

        return {
            "model": self.model,
            "max_tokens": tokens,
            "temperature": self.temperature,
            "messages": [{"role": "user", "content": prompt}],
            "system": self.system_message,
        }

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_if_exception_type(ANTHROPIC_EXCEPTIONS),
    )
    async def _make_request(
        self,
        index: str | None,
        prompt: str | None,
        tokens: int | None,
        repo_files: list[tuple[str, str]] | None,
    ) -> Any:
        """Processes Anthropic API responses and returns generated text."""
        if not ANTHROPIC_AVAILABLE:
            self._logger.error(
                "Cannot make request: Anthropic library is not available."
            )
            return index, self.placeholder

        if self.client is None:
            self._logger.error("Anthropic client is not properly initialized.")
            return index, self.placeholder

        try:
            prompt = await token_handler(self.config, index, prompt, tokens)
            parameters = await self._build_payload(prompt, tokens)

            async with self.rate_limit_semaphore:
                self._logger.info(f"Making request to Anthropic for '{index}'")
                response = await self.client.messages.create(**parameters)
                data = (
                    response.content[0].text
                    if hasattr(response, "content")
                    else str(response)
                )
                self._logger.info(
                    f"Response from Anthropic for '{index}': {data}"
                )
                return index, data

        except ANTHROPIC_EXCEPTIONS as e:
            self._logger.error(
                f"API Error processing request for '{index}': {e!r}"
            )
            raise  # Re-raise for retry decorator

        except Exception as e:
            self._logger.error(f"Unexpected error for '{index}': {e!r}")
            return index, self.placeholder



from pathlib import Path


class TreeGenerator:
    """
    Generates a directory tree structure for a code repository.
    """

    def __init__(
        self,
        repo_name: str,
        root_dir: Path,
        repo_url: str,
        max_depth: int,
    ):
        self.repo_name = repo_name
        self.root_dir = root_dir
        self.repo_url = repo_url
        self.max_depth = max_depth

    def generate(
        self,
        directory: Path,
        prefix: str = "",
        is_last: bool = True,
        depth: int = 0,
    ) -> str:
        """Builds a string representation of the directory structure."""
        if depth > self.max_depth:
            return ""

        children = sorted(directory.iterdir()) if directory.is_dir() else []
        children = list(children)

        if not children and directory.is_dir():
            return ""

        parts = [f"{prefix}{'└── ' if is_last else '├── '}{directory.name}"]

        for index, child in enumerate(children):
            child_prefix = prefix + ("    " if is_last else "│   ")
            if child_tree := self.generate(
                child,
                child_prefix,
                index == len(children) - 1,
                depth + 1,
            ):
                parts.append(child_tree)

        return self._format_tree(parts)

    def _format_tree(self, parts: list[str]) -> str:
        """Format the directory tree structure."""
        tree = "\n".join(parts)
        return tree.replace(self.root_dir.name, f"{self.repo_name}/")



"""Pydantic models and settings for the readme-ai package."""

from __future__ import annotations

import sys
from pathlib import Path
from typing import Literal, Self

from pydantic import (
    AnyHttpUrl,
    BaseModel,
    ConfigDict,
    Field,
    FilePath,
    NonNegativeFloat,
    PositiveInt,
    field_validator,
    model_validator,
)
from pydantic_extra_types.color import Color

from readmeai.config.constants import (
    BadgeStyleOptions,
    HeaderStyleOptions,
    ImageOptions,
    TocStyleOptions,
)
from readmeai.errors import GitValidationError
from readmeai.generators.banner import (
    convert_svg_to_html,
    generate_ascii_banner,
    generate_ascii_box_banner,
)
from readmeai.logger import get_logger
from readmeai.readers.git.providers import GitURL, parse_git_url
from readmeai.utils.file_handler import FileHandler
from readmeai.utils.file_resource import get_resource_path

_logger = get_logger(__name__)


class APISettings(BaseModel):
    """
    LLM API settings and parameters.
    """

    rate_limit: PositiveInt = Field(gt=0, le=25, description="API rate limit.")
    system_message: str = Field(description="LLM system prompt content field.")


class FileSettings(BaseModel):
    """
    File path resources for the readme-ai package.
    """

    ignore_list: str = Field(description="List of files to ignore.")
    languages: str = Field(description="Extension to language mappings.")
    parsers: str = Field(description="Common dependency file names.")
    prompts: str = Field(description="LLM API prompt templates.")
    tool_config: str = Field(description="Development tool configurations.")
    tooling: str = Field(description="Development tools and utilities.")
    shieldsio_icons: str = Field(description="Shields.io svg icon badges.")
    skill_icons: str = Field(description="Skill icon badges.")


class GitSettings(BaseModel):
    """
    User repository settings for a remote or local codebase.
    """

    repository: Path | str
    full_name: str | None = None
    host_domain: str | None = None
    host: str | None = None
    name: str = ""

    model_config = ConfigDict(extra="forbid")

    @field_validator("repository")
    def validate_repository(cls, value: Path | str) -> Path | str:
        """Validates the repository path or Git URL."""
        if isinstance(value, Path) or (
            isinstance(value, str)
            and Path(value).is_dir()
            and Path(value).exists()
        ):
            return value
        try:
            return str(GitURL.create(value).url)
        except ValueError as exc:
            raise GitValidationError(
                f"Invalid Git repository URL or path: {value}",
            ) from exc

    @model_validator(mode="after")
    def set_git_attributes(self) -> Self:
        """Parse and set Git repository attributes."""
        self.host_domain, self.host, self.name, self.full_name = parse_git_url(
            str(self.repository)
        )
        return self


class MarkdownSettings(BaseModel):
    """
    Markdown code template blocks for building the README.md file.
    """

    align: Literal["left", "center", "right"] = Field(
        default="center", description="align for markdown content."
    )
    badge_color: Color = Field(
        default_factory=lambda: Color("blue"),
        description="Badge color (https://www.w3.org/TR/SVG11/types.html#ColorKeywords)",
    )
    badge_style: BadgeStyleOptions = Field(
        default=BadgeStyleOptions.DEFAULT, description="Badge icon style type."
    )
    badges_tech_stack: str
    badges_tech_stack_text: str
    contribute: str
    emojis: bool = Field(
        default=False, description="Enable emoji prefixes for headers."
    )
    features: str = Field(
        default="❯ INSERT-PROJECT-FEATURES",
        description="Project feature markdown table content.",
    )
    header_style: str = Field(
        default=HeaderStyleOptions.CLASSIC,
        description="Header style for the project README.",
    )
    image: AnyHttpUrl | FilePath | str = Field(
        default=ImageOptions.BLUE,
        description="Project image URL or file path.",
    )
    image_width: str = Field(default="20%")
    overview: str = Field(default="❯ INSERT-PROJECT-OVERVIEW")
    placeholder: str = Field(default="<code>❯ REPLACE-ME</code>")
    project_index: str
    quickstart: str
    shieldsio_icons: str
    skill_icons: str
    slogan: str = Field(default="❯ INSERT-PROJECT-SLOGAN")
    tables: str = Field(default="❯ INSERT-PROJECT-TABLES")
    toc_style: str = Field(default=TocStyleOptions.BULLET)
    tree: str
    tree_depth: PositiveInt = Field(default=2, ge=1, le=5)

    model_config = ConfigDict(
        use_enum_values=True,
        arbitrary_types_allowed=True,
    )

    @field_validator("badge_color")
    def set_color(cls, value: str) -> str:
        """Field validator to set the badge color."""
        try:
            return Color(value).as_hex(format="long").lstrip("#")
        except ValueError:
            _logger.error(f"Invalid color provided: {value}", exc_info=True)
            return cls.model_fields["badge_color"].default


class ModelSettings(BaseModel):
    """
    LLM API model settings and parameters.
    """

    api: str
    base_url: str
    context_window: PositiveInt
    encoder: str
    host_name: AnyHttpUrl
    localhost: AnyHttpUrl
    model: str
    path: str
    temperature: NonNegativeFloat
    tokens: PositiveInt
    top_p: NonNegativeFloat


class Settings(BaseModel):
    """
    Pydantic settings model for the readme-ai package.
    """

    api: APISettings
    files: FileSettings
    git: GitSettings
    llm: ModelSettings
    md: MarkdownSettings

    model_config = ConfigDict(
        validate_assignment=True,
    )

    @model_validator(mode="after")
    def generate_banner(self) -> Self:
        """Generates the project banner based on the settings."""
        header_style = self.md.header_style.lower()
        if header_style == HeaderStyleOptions.ASCII.value:
            self.md.image = generate_ascii_banner(self.git.name)
            self.md.header_style = HeaderStyleOptions.ASCII
        elif header_style == HeaderStyleOptions.ASCII_BOX.value:
            self.md.image = generate_ascii_box_banner(self.git.name)
            self.md.header_style = HeaderStyleOptions.ASCII
        elif header_style == HeaderStyleOptions.SVG.value:
            self.md.image = convert_svg_to_html(
                self.git.name, f"{self.git.name}-banner.svg"
            )
            self.md.header_style = HeaderStyleOptions.SVG
        elif (
            header_style
            in [
                HeaderStyleOptions.CLASSIC.value,
                HeaderStyleOptions.COMPACT.value,
                HeaderStyleOptions.MODERN.value,
            ]
            or self.md.image == ImageOptions.LLM.value
        ):
            if header_style == HeaderStyleOptions.CLASSIC.value:
                self.md.header_style = HeaderStyleOptions.CLASSIC
                self.md.image_width = "30%"
            elif header_style == HeaderStyleOptions.COMPACT.value:
                self.md.header_style = HeaderStyleOptions.COMPACT
                self.md.image_width = "40%"
            elif header_style == HeaderStyleOptions.MODERN.value:
                self.md.header_style = HeaderStyleOptions.MODERN
                self.md.image_width = "30%"

        return self


class ConfigLoader:
    """
    Loads the configuration settings for the readme-ai package.
    """

    file_handler: FileHandler = FileHandler()
    config_file: str = "config.toml"
    module: str = "readmeai.config"
    submodule: str = "settings"

    def __init__(self) -> None:
        """Initialize ConfigLoader with the base configuration file."""
        if "-V" not in sys.argv and "--version" not in sys.argv:
            self._load_config()
            self._load_settings()

    def _load_config(self) -> Settings:
        """Loads the base configuration file."""
        file_path = get_resource_path(
            file_path=self.config_file,
            submodule=self.submodule,
        )
        config_dict = self.file_handler.read(file_path)
        self.config = Settings.model_validate(config_dict)
        return self.config

    def _load_settings(self) -> dict[str, dict]:
        """Loads all TOML config files from ./config/settings/*.toml."""
        settings = self.config.model_dump()

        for key, file_path in settings["files"].items():
            if file_path.endswith(".toml"):
                file_path = get_resource_path(file_path=file_path)
                config_dict = self.file_handler.read(file_path)
                settings[key] = config_dict
                setattr(self, key, config_dict)
                _logger.info(f"Configuration file loaded: {file_path}")

        return settings



import re
from typing import Any, ClassVar

from readmeai.config.constants import TocStyleOptions
from readmeai.templates.base import BaseTemplate


class TocTemplate(BaseTemplate):
    """
    Class variable for rendering the README.md Table of Contents.
    """

    TOC_TEMPLATES: ClassVar[dict] = {
        TocStyleOptions.BULLET: """## 🔗 Table of Contents\n\n{toc_items}\n---\n""",
        TocStyleOptions.FOLD: """<details><summary>Table of Contents</summary>\n\n{toc_items}\n</details>\n<hr>\n""",
        TocStyleOptions.NUMBER: """## 🔗 Table of Contents\n\n{toc_items}\n---\n""",
        TocStyleOptions.LINKS: """## 🔗 Quick Links\n\n{toc_items}\n---\n""",
        TocStyleOptions.ROMAN: """## 🔗 Table of Contents\n\n{toc_items}\n---\n""",
    }

    TOC_ITEM_TEMPLATES: ClassVar[dict[TocStyleOptions, str]] = {
        TocStyleOptions.BULLET: "- [{title}](#{anchor})\n",
        TocStyleOptions.FOLD: "- [{title}](#{anchor})\n",
        TocStyleOptions.NUMBER: "{number}. [{title}](#{anchor})\n",
        TocStyleOptions.LINKS: "- [{title}](#{anchor})\n",
        TocStyleOptions.ROMAN: "{number}. [{title}](#{anchor})\n",
    }

    def __init__(self, style: str = TocStyleOptions.BULLET.value) -> None:
        self.style = TocStyleOptions(style)

    @staticmethod
    def generate_anchor(title: str) -> str:
        """Generate an anchor link from the given title.
        # #-table-of-contents
        """
        emoji_pattern = re.compile(
            "["
            "\U0001f600-\U0001f64f"  # emoticons
            "\U0001f300-\U0001f5ff"  # symbols & pictographs
            "\U0001f680-\U0001f6ff"  # transport & map symbols
            "\U0001f1e0-\U0001f1ff"  # flags (iOS)
            "\U00002702-\U000027b0"
            "\U000024c2-\U0001f251"
            "]+",
            flags=re.UNICODE,
        )
        title = emoji_pattern.sub("", title)
        return title.lower().replace(" ", "-")

    def _generate_toc_items(
        self,
        sections: list[dict[str, Any]],
        level: int = 0,
        parent_number: str = "",
    ) -> str:
        """Generate Table of Contents items recursively."""
        toc = ""
        for index, section in enumerate(sections, start=1):
            if self.style == TocStyleOptions.ROMAN and level > 0:
                continue

            indent = "  " * level
            title = section["title"]
            anchor = self.generate_anchor(title)

            if self.style in {TocStyleOptions.NUMBER, TocStyleOptions.ROMAN}:
                if self.style == TocStyleOptions.ROMAN:
                    number = (
                        f"{parent_number}{self._int_to_roman(index)}"
                        if parent_number
                        else self._int_to_roman(index)
                    )
                else:
                    number = (
                        f"{parent_number}{index}"
                        if parent_number
                        else str(index)
                    )
                item = self.TOC_ITEM_TEMPLATES[self.style].format(
                    number=number,
                    title=title,
                    anchor=anchor,
                )
            else:
                item = self.TOC_ITEM_TEMPLATES[self.style].format(
                    title=title,
                    anchor=anchor,
                )

            toc += f"{indent}{item}"

            if (
                "subsections" in section
                and self.style != TocStyleOptions.ROMAN
            ):
                toc += self._generate_toc_items(
                    section["subsections"],
                    level + 1,
                    f"{number}."
                    if self.style == TocStyleOptions.NUMBER
                    else "",
                )

        return toc

    @staticmethod
    def _int_to_roman(number: int) -> str:
        """Convert an integer to a Roman numeral."""
        val = [1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1]
        syb = [
            "M",
            "CM",
            "D",
            "CD",
            "C",
            "XC",
            "L",
            "XL",
            "X",
            "IX",
            "V",
            "IV",
            "I",
        ]
        roman_num = ""
        i = 0
        while number > 0:
            for _ in range(number // val[i]):
                roman_num += syb[i]
                number -= val[i]
            i += 1
        return roman_num

    def render(self, data: dict[str, Any]) -> str:
        """Render Table of Contents based on the current style and data."""
        toc_items = self._generate_toc_items(data["sections"])
        template = self.TOC_TEMPLATES[self.style]
        return template.format(toc_items=toc_items)

    @staticmethod
    def get_toc_template(template: str) -> str:
        """Get the Table of Contents template for the given style."""
        return TocTemplate.TOC_TEMPLATES[TocStyleOptions(template)]



from abc import ABC, abstractmethod

from readmeai.logger import get_logger


class BaseFileParser(ABC):
    """
    Abstract base class for dependency file parsers.
    """

    def __init__(self) -> None:
        self._logger = get_logger(__name__)

    @abstractmethod
    def parse(self, content: str) -> list[str]:
        """Parses content of file and return list of dependencies."""
        ...

    def log_error(self, message: str):
        """Logs error message when parsing fails."""
        self._logger.error(f"Error parsing dependency file {message}")

    def handle_parsing_error(self, error: Exception) -> list[str]:
        """Standardized error handling for parsing exceptions."""
        self.log_error(repr(error))
        return []


class DefaultParser(BaseFileParser):
    """
    Default parser for unknown file types.
    """

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> list[str]:
        """Returns an empty list for unknown file types."""
        return []



"""Parsers for npm related dependency files i.e. package.json"""

import json

from readmeai.parsers.base import BaseFileParser


class PackageJsonParser(BaseFileParser):
    """
    Parser for package.json dependency files.
    """

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> list[str]:
        """Returns a list of dependencies parsed from a json file."""
        try:
            data = json.loads(content)
            package_names = []
            for section in [
                "dependencies",
                "devDependencies",
                "peerDependencies",
            ]:
                if section in data:
                    package_names.extend(data[section].keys())
            return package_names

        except json.JSONDecodeError as exc:
            return self.handle_parsing_error(exc)



from collections import Counter
from pathlib import Path

from readmeai.config.settings import ConfigLoader
from readmeai.ingestion.models import FileContext
from readmeai.logger import get_logger
from readmeai.parsers.factory import ParserFactory
from readmeai.preprocessor.document_cleaner import DocumentCleaner
from readmeai.preprocessor.file_filter import is_excluded

_logger = get_logger(__name__)


class FileProcessor:
    """
    File processor class to process files in a repository.
    """

    def __init__(self, config: ConfigLoader):
        self.config = config
        self.document_cleaner = DocumentCleaner()
        self.ignore_list = config.ignore_list.get("ignore_list", [])
        self.language_names = config.languages.get("language_names", {})

    def process_files(self, repo_path: Path) -> list[FileContext]:
        """Generate file info for the given repository path."""
        return [
            self._create_file_context(file_path, repo_path)
            for file_path in repo_path.rglob("*")
            if file_path.is_file()
            and not is_excluded(self.ignore_list, file_path, repo_path)
        ]

    def count_languages(
        self, file_contexts: list[FileContext]
    ) -> dict[str, int]:
        """Count the occurrences of each language."""
        return dict(Counter(file.ext for file in file_contexts if file.ext))

    def extract_dependencies(
        self, file_contexts: list[FileContext]
    ) -> list[str]:
        """Extract all dependencies from file contexts."""
        return list(
            set().union(*(file.dependencies for file in file_contexts))
        )

    def _create_file_context(
        self, file_path: Path, repo_path: Path
    ) -> FileContext:
        """Create a file context object for the given file path."""
        relative_path = str(file_path.relative_to(repo_path))
        content = file_path.read_text(errors="ignore")
        file_ext = file_path.suffix.lstrip(".")
        return FileContext(
            path=relative_path,
            name=file_path.name,
            ext=file_ext,
            content=self.document_cleaner.clean(content),
            language=self._map_language(file_ext, file_path.name),
            dependencies=self._parse_dependencies(relative_path, content),
        )

    def _map_language(self, file_ext: str, file_name: str) -> str:
        """Map the file extension to the programming language name."""
        return self.language_names.get(file_ext, file_name)

    def _parse_dependencies(self, file_path: str, content: str) -> list[str]:
        """Parse dependencies from the file content."""
        if parser := ParserFactory.create_parser(file_path):
            try:
                return parser.parse(content) or []
            except Exception as e:
                _logger.error(
                    f"Error parsing dependency file {file_path}: {e}"
                )
        return []



"""Utilities for handling tokens in the LLM model."""

import structlog
from tiktoken import Encoding, get_encoding

from readmeai.config.settings import Settings
from readmeai.logger import get_logger

_encoding_cache = {}

_logger = get_logger(__name__)


def _set_encoding_cache(encoding_name: str) -> Encoding:
    """Set the encoding cache for a specific encoding."""
    if encoding_name not in _encoding_cache:
        _encoding_cache[encoding_name] = get_encoding(encoding_name)
    return _encoding_cache[encoding_name]


async def token_handler(
    config: Settings,
    index: str,
    prompt: str,
    tokens: int,
) -> str:
    """Handle token count for the prompt."""
    encoder = config.llm.encoder
    max_count = config.llm.context_window
    token_count = count_tokens(prompt, encoder)

    structlog.contextvars.bind_contextvars(
        token_count=token_count,
        max_count=max_count,
        index=index,
    )

    if token_count > max_count:
        _logger.debug(
            f"Truncating '{index}' prompt: {token_count} > {max_count} tokens!",
        )
        prompt = truncate_tokens(encoder, prompt, tokens)

    return prompt


def count_tokens(text: str, encoder: str) -> int:
    """Return the number of tokens in a text string."""
    try:
        encoding = _set_encoding_cache(encoder)
        token_count = len(encoding.encode(text, disallowed_special=()))

    except (UnicodeEncodeError, ValueError) as exc:
        _logger.error(
            f"Error counting tokens for '{text}' with {encoder}: {exc}",
        )
        token_count = 0

    return token_count


def truncate_tokens(encoding: str, text: str, max_count: int) -> str:
    """Truncate a text string to a maximum number of tokens."""
    if not text:
        return text
    try:
        encoder = _set_encoding_cache(encoding)
        token_count = len(encoder.encode(text))
        if token_count <= max_count:
            return text
        char_total = len(text)
        chars_per_token = char_total / token_count
        truncated_total = int(chars_per_token * max_count)
        return text[:truncated_total]

    except Exception as exc:
        _logger.error(f"Error truncating tokens for '{text}': {exc}")
        return text


def update_max_tokens(
    max_tokens: int,
    prompt: str,
    target: str = "Hello!",
) -> int:
    """Adjust the maximum number of tokens based on the specific prompt."""
    is_valid_prompt = prompt.strip().startswith(target.strip())
    return max_tokens if is_valid_prompt else max_tokens // 2



"""Parser for *.properties configuration files."""

import re

from .base import BaseFileParser


class PropertiesParser(BaseFileParser):
    def __init__(self):
        super().__init__()
        super().__init__()
        self.common_tech_keywords = {
            "spring",
            "gradle",
            "maven",
            "java",
            "kotlin",
            "python",
            "node",
            "react",
            "angular",
            "vue",
            "docker",
            "kubernetes",
            "aws",
            "azure",
            "gcp",
            "android",
            "ios",
            "swift",
            "flutter",
            "django",
            "flask",
            "fastapi",
            "express",
            "nest",
            "tensorflow",
            "pytorch",
            "pandas",
            "numpy",
            "scikit",
            "hadoop",
            "spark",
            "kafka",
            "redis",
            "mongodb",
            "postgresql",
            "mysql",
            "oracle",
            "elasticsearch",
            "graphql",
            "rest",
            "grpc",
            "protobuf",
            "openapi",
            "swagger",
            "selenium",
            "terraform",
            "ansible",
            "helm",
            "git",
            "svn",
            "cicd",
            "jenkins",
            "circleci",
            "gitlab",
            "github",
            "bitbucket",
            "travis",
            "webpack",
            "babel",
            "typescript",
            "ruby",
            "rails",
            "scala",
            "clojure",
            "erlang",
            "elixir",
            "phoenix",
            "laravel",
            "svelte",
            "webassembly",
            "rust",
            "go",
            "julia",
            "csharp",
            ".net",
            "asp.net",
            "qml",
            "unity",
            "unreal",
            "cryengine",
            "vulkan",
            "opengl",
            "threejs",
            "blender",
            "matlab",
            "octave",
            "fortran",
            "r",
            "tableau",
            "powerbi",
            "qlik",
            "sap",
            "salesforce",
            "dynamics",
            "stripe",
            "paypal",
            "braintree",
            "rabbitmq",
            "zeromq",
            "socket.io",
            "websockets",
            "webrtc",
            "oauth",
            "jwt",
            "saml",
            "openid",
            "hashicorp",
            "cassandra",
            "neo4j",
            "arangodb",
            "couchbase",
            "dynamodb",
            "memcached",
            "sphinx",
            "solr",
            "lucene",
            "next.js",
            "nuxt.js",
            "remix",
            "gatsby",
            "emotion",
            "styled-components",
            "tailwindcss",
            "bootstrap",
            "material-ui",
            "ant-design",
            "chakra-ui",
            "quasar",
            "primefaces",
            "vuetify",
            "fivem",
            "qt",
            "wxwidgets",
            "tkinter",
            "xamarin",
            "react-native",
            "electron",
            "cordova",
            "ionic",
            "nativescript",
            "appium",
            "xstate",
            "rxjs",
            "mobx",
            "redux",
            "vuex",
            "recoil",
            "pinia",
            "akka",
            "vert.x",
            "play",
            "quarkus",
            "micronaut",
            "spring-boot",
            "netty",
            "kotlinx.coroutines",
            "jetpack",
            "room",
            "workmanager",
            "compose",
            "pulumi",
            "boto3",
            "sdkman",
            "jvm",
            "graalvm",
            "openjdk",
            "v8",
            "spidermonkey",
            "webgl",
            "d3.js",
            "highcharts",
            "amcharts",
            "echarts",
            "plotly",
            "vega",
            "kendo-ui",
        }

    def parse(self, content: str) -> list[str]:
        lines = content.split("\n")
        dependencies = set()

        for line in lines:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            dependencies.update(self._extract_from_line(line))

        return sorted(list(dependencies))

    def _clean_word(self, word: str) -> str:
        word = word.lower()
        word = re.sub(
            r"^(lib|org|com|net|io|api|sdk|cli|ui|app|service|utils?|helpers?|core|base|impl|test|dev|prod)\.?",
            "",
            word,
        )
        word = re.sub(
            r"(version|tool|module|plugin|dependency|lib|framework|platform)s?$",
            "",
            word,
        )
        return word.strip()

    def _extract_from_line(self, line: str) -> set[str]:
        extracted = set()

        if "=" in line:
            key, value = line.split("=", 1)
            extracted.update(self._extract_words(key))
            extracted.update(self._extract_words(value))

            # Handle version-specific dependencies
            if "version" in key.lower():
                dep_name = re.sub(
                    r"version.*$", "", key, flags=re.IGNORECASE
                ).strip()
                extracted.add(f"{dep_name}: {value}")
        else:
            extracted.update(self._extract_words(line))

        return extracted

    def _extract_words(self, text: str) -> set[str]:
        words = {text.lower()}
        parts = re.split(r"[._-]", text)
        words.update(part.lower() for part in parts)

        for i in range(len(parts)):
            for j in range(i + 1, len(parts) + 1):
                words.add(".".join(parts[i:j]).lower())

        # Split camelCase
        for part in parts:
            words.update(self._split_camel_case(part))

        return words

    def _filter_technologies(self, words: set[str]) -> set[str]:
        return {
            word
            for word in words
            if word in self.common_tech_keywords or len(word) > 1
        }

    def _split_camel_case(self, text: str) -> set[str]:
        words = re.findall(
            r"[A-Z]?[a-z]+|[A-Z]{2,}(?=[A-Z][a-z]|\d|\W|$)|\d+", text
        )
        return {word.lower() for word in words}



"""Methods to build SVG badges for the README using shields.io icons."""

import colorsys

from readmeai.config.constants import BadgeStyleOptions
from readmeai.config.settings import Settings
from readmeai.readers.git.providers import GitHost
from readmeai.utils.file_handler import FileHandler
from readmeai.utils.file_resource import get_resource_path

_comment = "<!-- default option, no dependency badges. -->\n"
_package = "readmeai.generators"
_submodule = "svg"


def build_default_badges(
    config: Settings,
    full_name: str,
    host: str,
) -> str:
    """Build metadata badges using shields.io."""
    return config.md.shieldsio_icons.format(
        host=host,
        full_name=full_name,
        badge_color=config.md.badge_color,
        badge_style=config.md.badge_style,
    )


def build_badges_tech_stack(
    dependencies: list[str],
    icons: dict[str, str],
    style: str,
) -> str:
    """Build HTML badges for project dependencies."""
    badges = [
        icons[str(dependency).lower()]
        for dependency in dependencies
        if str(dependency).lower() in icons
    ]
    badges = sort_badges(badges)
    badges = [badge[0].format(style) for badge in badges]
    return format_badges(badges)


def format_badges(badges: list[str]) -> str:
    """Format SVG badge icons as HTML."""
    total = len(badges)
    if badges is None or total == 0:
        return ""

    badges_per_line = total if total < 9 else (total // 2) + (total % 2)

    lines = []
    for i in range(0, total, badges_per_line):
        line = "\n\t".join(
            [
                f'<img src="{badge}" alt="{badge.split("/badge/")[1].split("-")[0]}">'
                for badge in badges[i : i + badges_per_line]
            ],
        )
        lines.append(
            f"{line}\n\t<br>" if i + badges_per_line < total else f"{line}\n",
        )

    return "\n\t".join(lines)


def hex_to_hls(hex_color: str) -> tuple[float, float, float]:
    """Converts a hex color to HLS."""
    hex_color = hex_color.lstrip("#")
    rgb = tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))
    return colorsys.rgb_to_hls(rgb[0] / 255, rgb[1] / 255, rgb[2] / 255)


def sort_badges(badges: list[tuple[str, str]]) -> list[str]:
    """Sorts badges by color and then by name."""
    badges = [(badge[0], str(badge[1])) for badge in badges]
    badges = list(set(badges))
    return sorted(
        badges,
        key=lambda b: (*hex_to_hls(b[1]), b[0])
        if b[1]
        else (float("inf"), 0, 0, ""),
    )


def shieldsio_icons(
    conf: Settings,
    dependencies: list,
    full_name: str,
    git_host: str,
) -> tuple[str, str]:
    """Generates badges for the README using shields.io icons."""
    icons_path = get_resource_path(
        conf.files.shieldsio_icons,
        _package,
        _submodule,
    )
    icons_dict = FileHandler().read(icons_path)

    default_icons = build_default_badges(conf, full_name, git_host)

    badges_tech_stack = build_badges_tech_stack(
        dependencies,
        icons_dict,
        conf.md.badge_style,
    )
    badges_tech_stack = conf.md.badges_tech_stack.format(
        align=conf.md.align,
        badges_tech_stack=badges_tech_stack,
    )

    if (
        conf.md.badge_style == BadgeStyleOptions.DEFAULT.value
        and git_host != GitHost.LOCAL.name
    ):
        conf.md.badges_tech_stack_text = _comment
        return (
            default_icons,
            _comment,
        )

    if git_host == GitHost.LOCAL.name:
        return (
            "<!-- local repository, no metadata badges. -->",
            badges_tech_stack,
        )

    return default_icons, badges_tech_stack


def skill_icons(conf: Settings, dependencies: list) -> str:
    """Generate 'skill-icons' badge set for the README header.
    Source: https://github.com/tandpfun/skill-icons
    """
    dependencies.extend(["md"])

    icons_path = get_resource_path(
        conf.files.skill_icons,
        _package,
        _submodule,
    )
    icons_dict = FileHandler().read(icons_path)

    icons = [
        icon for icon in icons_dict["icons"]["names"] if icon in dependencies
    ]
    formatted_icons = icons_dict["url"]["base_url"] + ",".join(icons)

    if conf.md.badge_style == "skills-light":
        formatted_icons = f"{formatted_icons}&theme=light"

    conf.md.skill_icons = conf.md.skill_icons.format(formatted_icons)

    return conf.md.badges_tech_stack.format(
        align=conf.md.align,
        badges_tech_stack=conf.md.skill_icons,
    )



from pathlib import Path

from readmeai.config.settings import BadgeStyleOptions, ConfigLoader
from readmeai.generators import badges, emojis, tables, tree
from readmeai.ingestion.models import RepositoryContext
from readmeai.readers.git.providers import GitHost
from readmeai.templates.header import HeaderTemplate
from readmeai.templates.quickstart import QuickStartBuilder
from readmeai.templates.table_of_contents import TocTemplate


class MarkdownBuilder:
    """
    Builds each section of the README Markdown file.
    """

    def __init__(
        self,
        config_loader: ConfigLoader,
        repo_context: RepositoryContext,
        file_summaries: list[tuple[str, str]],
        temp_dir: str,
    ):
        self.config_loader = config_loader
        self.config = config_loader.config
        self.deps = repo_context.dependencies
        self.repo_context = repo_context
        self.summaries = file_summaries
        self.temp_dir = Path(temp_dir)
        self.md = self.config.md
        self.git = self.config.git
        self.repo_url = (
            self.git.repository
            if self.git.host_domain != GitHost.LOCAL.name.lower()
            else f"../{self.git.name}"
        )
        self.header_template = HeaderTemplate(self.md.header_style)
        self.toc_template = TocTemplate(self.md.toc_style)

    @property
    def header_and_badges(self) -> str:
        """Generates the README header section."""
        if BadgeStyleOptions.SKILLS.value not in self.md.badge_style:
            md_shields, md_badges = badges.shieldsio_icons(
                self.config,
                self.deps,
                str(self.git.full_name),
                str(self.git.host),
            )
        else:
            md_shields = (
                "<!-- Shields.io badges disabled, using skill icons. -->"
            )
            md_badges = badges.skill_icons(self.config, self.deps)

        header_data = {
            "align": self.md.align,
            "image": self.md.image,
            "image_width": self.md.image_width,
            "repo_name": self.git.name.upper()
            if self.git.name
            else self.md.placeholder,
            "slogan": self.md.slogan,
            "shields_icons": md_shields,
            "badges_tech_stack": md_badges,
            "badges_tech_stack_text": self.md.badges_tech_stack_text,
        }
        return self.header_template.render(header_data)

    @property
    def table_of_contents(self) -> str:
        """Generates the README Table of Contents section."""
        headings = {
            "sections": [
                {"title": "📍 Overview"},
                {"title": "👾 Features"},
                {
                    "title": "📁 Project Structure",
                    "subsections": [
                        {"title": "📂 Project Index"},
                    ],
                },
                {
                    "title": "🚀 Getting Started",
                    "subsections": [
                        {"title": "☑️ Prerequisites"},
                        {"title": "⚙️ Installation"},
                        {"title": "🤖 Usage"},
                        {"title": "🧪 Testing"},
                    ],
                },
                {"title": "📌 Project Roadmap"},
                {"title": "🔰 Contributing"},
                {"title": "🎗 License"},
                {"title": "🙌 Acknowledgments"},
            ],
        }
        return self.toc_template.render(headings)

    @property
    def file_summaries(self) -> str:
        """Generates markdown tables that store repository module summaries."""
        formatted_summaries = tables.format_code_summaries(
            self.md.placeholder,
            self.summaries,
        )
        return tables.generate_nested_module_tables(
            formatted_summaries,
            self.git.full_name,
            self.repo_url,
        )

    @property
    def tree(self) -> str:
        """Generates the README directory tree structure."""
        project_tree = tree.TreeGenerator(
            repo_name=self.git.name,
            root_dir=self.temp_dir,
            repo_url=self.repo_url,
            max_depth=self.md.tree_depth,
        ).generate(self.temp_dir)
        return self.md.tree.format(project_tree)

    @property
    def quickstart_guide(self) -> str:
        """Generates the README Getting Started section."""
        return QuickStartBuilder(self.config_loader, self.repo_context).build()

    @property
    def contributing_guide(self) -> str:
        """Generates the README Contributing section."""
        return self.md.contribute.format(
            host=self.git.host,
            host_domain=self.git.host_domain,
            full_name=self.git.full_name,
            repo_name=self.git.name,
            repo_url=self.repo_url,
        )

    def build(self) -> str:
        """Builds each section of the README.md file."""
        readme_md_contents = [
            self.header_and_badges,
            self.table_of_contents,
            self.md.overview,
            self.md.features,
            self.tree,
            self.md.project_index,
            self.file_summaries,
            self.quickstart_guide,
            self.contributing_guide,
        ]

        if self.config.md.emojis is False:
            readme_md_contents = emojis.remove_emojis(readme_md_contents)

        return "\n".join(readme_md_contents)



from dataclasses import dataclass, field

from readmeai.config.settings import ConfigLoader
from readmeai.ingestion.models import QuickStart
from readmeai.logger import get_logger

_logger = get_logger(__name__)


@dataclass
class QuickStartGenerator:
    """
    Build the data model for generating the 'Quickstart' instructions.
    """

    config: ConfigLoader
    tools: dict = field(init=False)
    language_names: dict = field(init=False)
    default_commands: dict = field(init=False)

    def __post_init__(self):
        self.tools = self.config.tool_config
        self.language_names = self.config.languages.get("language_names", {})
        self.default_commands = self.tools.get("default", {})

    def generate(
        self,
        language_counts: dict[str, int],
        metadata: dict[str, dict[str, str]],
    ) -> QuickStart:
        """Get any relevant commands for the Quickstart instructions."""
        try:
            primary_language = self._get_primary_language(language_counts)
            quickstart = QuickStart(
                primary_language=primary_language,
                language_counts=language_counts,
                package_managers=metadata.get("package_managers", {}),
                containers=metadata.get("containers", {}),
            )
            self._generate_commands(quickstart, primary_language)
            return quickstart

        except Exception as e:
            _logger.error(
                f"Error generating QuickStart setup: {e}", exc_info=True
            )
            return QuickStart()

    def _get_primary_language(self, counts: dict[str, int]) -> str | None:
        """Determine the primary language of the repository."""
        if not counts:
            return None
        counts = {k: v for k, v in counts.items() if k not in ("yaml", "yml")}
        primary_lang = max(counts, key=counts.get)
        return self.language_names.get(
            primary_lang, self.language_names.get("default")
        )

    def _generate_commands(
        self, quickstart: QuickStart, primary_language: str
    ) -> None:
        """Generate install, usage, and test commands."""
        command_types = ["install", "usage", "test"]
        tool_types = ["package_managers", "containers"]
        for cmd_type in command_types:
            commands: list[str] = []
            for tool_type in tool_types:
                tools = getattr(quickstart, tool_type)
                commands.extend(
                    filter(
                        None,
                        [
                            self._format_command(
                                tool_name,
                                primary_language,
                                file_path,
                                cmd_type,
                                tool_type,
                            )
                            for tool_name, file_path in tools.items()
                        ],
                    )
                )
            setattr(quickstart, f"{cmd_type}_commands", "\n".join(commands))

    def _format_command(
        self,
        tool_name: str,
        primary_language: str,
        file_path: str,
        cmd_type: str,
        tool_type: str,
    ) -> str | None:
        """Format a command for the Quickstart instructions."""
        config = (
            self.tools.get(primary_language.lower(), {})
            .get(tool_type, {})
            .get(tool_name, {})
        ) or self.tools.get(tool_type, {}).get(tool_name, {})

        cmd = config.get(cmd_type, self.default_commands.get(cmd_type))
        if not cmd:
            return None

        if cmd_type == "install" and tool_type == "containers":
            cmd = cmd.replace("{image_name}", self.config.config.git.full_name)
        elif cmd_type in {"install", "test"}:
            cmd = cmd.replace("{file}", file_path)
        elif cmd_type == "usage":
            cmd = cmd.replace("{executable}", self.config.config.git.name)
        return f"""
**Using `{tool_name}`** &nbsp; [<img align="center" src="{config.get('shield', '')}" />]({config.get('website', '')})

```sh
❯ {cmd}
```
"""



"""Utility functions cleaning and formatting LLM API responses."""

import re


def fix_markdown_table_rows(md_table: str) -> str:
    """Format a Markdown table with feature and description columns."""
    lines = md_table.split("||")

    formatted_md_table = (
        "| Feature | Description |\n|---------|-------------|\n"
    )

    for line in lines[2:]:
        clean_line = line.strip("|")
        parts = clean_line.split("|")

        if len(parts) >= 3:
            feature = parts[1].strip()
            description = parts[2].strip()
            formatted_row = f"| {feature} | {description} |\n"
            formatted_md_table += formatted_row

    return formatted_md_table


def format_markdown_table(text: str) -> str:
    """
    Pattern to match a Markdown table. Looks for a
    header row with at least two columns, followed by
    a separator row, and then one or more data rows.
    This version is designed to be more robust in removing
    text around the markdown table.
    """
    if "REPLACE-ME</code>" in text:
        return text

    pattern = (
        r"(?:.*\n)*(\|.*\|.*\n\|[-: ]+\|[-: ]+\|.*\n(?:\|.*\|.*\n)*)(?:.*\n)*"
    )
    match = re.search(pattern, text, re.DOTALL)
    return match[1].strip() if match else ""


def process_markdown(text):
    """Remove uneven Markdown syntax while preserving valid formatting."""
    # Remove extra asterisks at the end of lines
    text = re.sub(r"\*+$", "", text, flags=re.MULTILINE)

    # Remove unmatched bullets or hyphens at the beginning of lines
    text = re.sub(r"^[\s]*[-*]\s+", "", text, flags=re.MULTILINE)

    # Preserve valid bold and italic formatting
    # This regex handles nested bold and italic formatting
    text = re.sub(
        r"\*{1,2}(?P<content>[^*\n]+(?:\*{1,2}[^*\n]+\*{1,2}[^*\n]+)*)\*{1,2}",
        lambda m: m.group(0)
        if m.group(0).count("*") % 2 == 0
        else m.group(0)[1:-1],
        text,
    )

    # Remove standalone asterisks or underscores
    text = re.sub(r"(?<!\*)\*(?!\*)|(?<!_)_(?!_)", "", text)

    return text.strip()


def process_text(text: str) -> str:
    """Format and clean generated text from the LLM."""
    # Dynamically remove all text before and including the first colon if any exist
    text = re.sub(r"^[^:]*:\s*", "", text)

    # Remove any text before and including "**:"
    text = re.sub(r"\*\*:\s*", "", text, flags=re.DOTALL)

    # Remove single and double quotes that are missing their closing counterpart
    text = re.sub(r"['\"](.*?)$", r"\1", text)
    text = re.sub(r"^(.*?)['\"]", r"\1", text)

    # Remove specific pattern and rephrase
    text = re.sub(
        r"\*\*Code Summary:\*\*\s*(.*?)\s*provides functions to",
        r"Provides functions to",
        text,
        flags=re.DOTALL,
    )
    # Remove single and double quotes around any text
    text = re.sub(r"(?<!\w)['\"](.*?)['\"](?!\w)", r"\1", text)

    # Remove newlines and tabs
    text = text.replace("\n", "").replace("\t", "")

    # Remove non-letter characters from the beginning of the string
    text = re.sub(r"^[^a-zA-Z]*", "", text)

    # Remove extra white space around punctuation except for '('
    text = re.sub(r"\s*([)'.!,?;:])(?!\.\s*\w)", r"\1", text)

    # Remove extra white space before opening parentheses
    text = re.sub(r"(\()\s*", r"\1", text)

    # Replace multiple consecutive spaces with a single space
    text = re.sub(r" +", " ", text)

    # Remove extra white space around hyphens
    text = re.sub(r"\s*-\s*", "-", text)

    # Specifically target and remove trailing special characters like asterisks
    text = re.sub(r"\*+$", "", text)

    text = text.strip()

    # Ensure the first letter is capitalized if it's alphabetic
    if text and not text[0].isupper() and text[0].isalpha():
        text = text[0].upper() + text[1:]

    return text


def remove_quotes(text: str) -> str:
    """Remove quotes from a string if they exist."""
    if not text or len(text) < 2:
        return text
    quote_chars = ("'", '"', "`")
    return (
        text[1:-1] if text[0] == text[-1] and text[0] in quote_chars else text
    )



from importlib import resources
from pathlib import Path

from readmeai.errors import FileReadError
from readmeai.logger import get_logger

_logger = get_logger(__name__)


def get_resource_path(
    file_path: str,
    module: str = "readmeai.config",
    submodule: str = "settings",
) -> Path:
    """Retrieves the path to a resource file within the package.

    Tries importlib.resources first, falls back to pkg_resources.
    """
    try:
        return resources.files(module).joinpath(submodule, file_path)

    except (TypeError, FileNotFoundError) as exc:
        _logger.error(
            f"Error loading resource file via importlib.resources: {exc}"
        )

    try:
        import pkg_resources

        submodule = submodule.replace(".", "/")
        return Path(
            pkg_resources.resource_filename(
                module,
                f"{submodule}/{file_path}",
            ),
        ).resolve()

    except Exception as exc:
        raise FileReadError(
            "Error loading resource file via pkg_resources",
            f"{submodule}/{file_path}",
        ) from exc



from typing import Annotated, Any

from pydantic import BaseModel, Field, StringConstraints


class QuickStart(BaseModel):
    """
    Instructions for installation, usage, and testing a repository.
    """

    primary_language: str | None = None
    language_counts: dict[str, int] = Field(default_factory=dict)
    package_managers: dict[str, str] = Field(default_factory=dict)
    containers: dict[str, str] = Field(default_factory=dict)
    install_commands: str = ""
    usage_commands: str = ""
    test_commands: str = ""


class FileContext(BaseModel):
    """
    FileContext model for storing file information.
    """

    path: str
    name: str
    ext: str
    content: str
    language: Annotated[str, StringConstraints(to_lower=True)]
    dependencies: Annotated[list[str], Field(default_factory=list)]


class RepositoryContext(BaseModel):
    """
    RepositoryContext model for storing repository information
    """

    files: list[FileContext]
    dependencies: list[str]
    languages: list[str]
    language_counts: dict[str, int]
    metadata: dict[str, Any] = Field(default_factory=dict)
    quickstart: QuickStart = Field(default_factory=QuickStart)



from typing import ClassVar

from readmeai.config.constants import LLMService
from readmeai.config.settings import ConfigLoader
from readmeai.errors import UnsupportedServiceError
from readmeai.ingestion.models import RepositoryContext
from readmeai.models.anthropic import AnthropicHandler
from readmeai.models.base import BaseModelHandler
from readmeai.models.gemini import GeminiHandler
from readmeai.models.offline import OfflineHandler
from readmeai.models.openai import OpenAIHandler


class ModelFactory:
    """
    Factory class for creating LLM API handler instances.
    """

    _model_map: ClassVar[dict] = {
        LLMService.ANTHROPIC: AnthropicHandler,
        LLMService.GEMINI.value: GeminiHandler,
        LLMService.OLLAMA.value: OpenAIHandler,
        LLMService.OPENAI.value: OpenAIHandler,
        LLMService.OFFLINE.value: OfflineHandler,
    }

    @staticmethod
    def get_backend(
        config: ConfigLoader, context: RepositoryContext
    ) -> BaseModelHandler:
        """
        Returns the appropriate LLM API handler based on CLI input.
        """
        llm_service = ModelFactory._model_map.get(config.config.llm.api)

        if llm_service is None:
            raise UnsupportedServiceError(
                f"Unsupported LLM service provided: {config.config.llm.api}",
            )

        return llm_service(config, context)



import fnmatch

from readmeai.config.settings import ConfigLoader
from readmeai.ingestion.models import FileContext


class MetadataExtractor:
    """
    Metadata extractor class to extract metadata from file contexts.
    """

    def __init__(self, config: ConfigLoader) -> None:
        self.config = config

    def extract_metadata(
        self, file_contexts: list[FileContext]
    ) -> dict[str, dict[str, str]]:
        """Extract metadata from file contexts, ensuring valid string values."""
        metadata_categories = {
            "cicd": "cicd",
            "containers": "containers",
            "documentation": "documentation",
            "package_managers": "package_managers",
        }
        return {
            category: self._convert_to_string(
                self._detect_tools(
                    file_contexts, self.config.tooling.get(tool_config, {})
                )
            )
            for category, tool_config in metadata_categories.items()
        }

    def _detect_tools(
        self, files: list[FileContext], tool_definitions: dict[str, list[str]]
    ) -> dict[str, list[str]]:
        """Detect tools based on file patterns."""
        detected_tools: dict[str, list[str]] = {}

        for file in files:
            for tool, patterns in tool_definitions.items():
                if matching_files := [
                    file.path
                    for pattern in patterns
                    if self._match_file_pattern(file.path, pattern)
                ]:
                    if tool not in detected_tools:
                        detected_tools[tool] = []
                    detected_tools[tool].extend(matching_files)

        return detected_tools

    def _match_file_pattern(self, file_path: str, pattern: str) -> bool:
        """
        Match a file path against a pattern.
        """
        if pattern.endswith("*"):
            return file_path.startswith(pattern.rstrip("*"))
        elif pattern.startswith("*"):
            return file_path.endswith(pattern[1:])
        elif "*" in pattern:
            return fnmatch.fnmatch(file_path, pattern)
        return file_path.endswith(pattern)

    def _convert_to_string(
        self, tool_files: dict[str, list[str]]
    ) -> dict[str, str]:
        """Convert detected tools list of file paths into a single string.

        - If multiple files are detected, join into a comma-separated string.
        """
        return {tool: ", ".join(files) for tool, files in tool_files.items()}



"""Custom structured logging implementation using structlog."""

import json
import logging
from collections.abc import MutableMapping
from typing import Annotated, Any

import structlog
from pydantic import BaseModel, Field, StringConstraints
from structlog.processors import CallsiteParameter
from structlog.stdlib import ProcessorFormatter
from structlog.types import Processor


class LoggingConfig(BaseModel):
    """
    Logging default configuration.
    """

    log_level: Annotated[str, StringConstraints(to_upper=True)] = "INFO"
    indent: Annotated[Any, Field(ge=2, le=50)] = 4
    pad_event: int = Field(default=20, ge=0, le=50)
    use_json: bool = True


class Logger:
    """
    Structured logging via structlog for the readme-ai package.
    """

    def __init__(self) -> None:
        self.log_config: LoggingConfig = LoggingConfig()
        self.log_level: str = self.log_config.log_level
        self.indent: int = self.log_config.indent
        self.pad_event: int = self.log_config.pad_event
        self.use_json: bool = self.log_config.use_json

    def configure_logger(self) -> None:
        """Configure the logger."""
        processors = self.get_processors()
        renderer = self.get_renderer()

        structlog.configure(
            processors=[*processors, renderer],
            context_class=dict,
            wrapper_class=structlog.stdlib.BoundLogger,
            logger_factory=structlog.stdlib.LoggerFactory(),
            cache_logger_on_first_use=True,
        )

        root_logger = logging.getLogger()
        handler = logging.StreamHandler()

        formatter = ProcessorFormatter(
            processor=renderer,
            foreign_pre_chain=processors,
        )

        handler.setFormatter(formatter)
        root_logger.addHandler(handler)
        root_logger.setLevel(self.log_config.log_level)

    def get_processors(self) -> list[Processor]:
        """Return list of structlog processors."""
        return [
            structlog.contextvars.merge_contextvars,
            structlog.stdlib.add_log_level,
            structlog.stdlib.add_logger_name,
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.CallsiteParameterAdder(
                [
                    CallsiteParameter.FILENAME,
                    CallsiteParameter.FUNC_NAME,
                    CallsiteParameter.LINENO,
                    # CallsiteParameter.THREAD_NAME,
                    # CallsiteParameter.PROCESS_NAME,
                ]
            ),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            self.format_message_processor,
        ]

    def get_renderer(self) -> Processor:
        """Sets the logging renderer to JSON or Console."""
        if self.use_json:
            return structlog.processors.JSONRenderer(indent=self.indent)
        else:
            return structlog.dev.ConsoleRenderer(pad_event=self.pad_event)

    @staticmethod
    def format_message_processor(
        logger: structlog.BoundLogger,
        method_name: str,
        event_dict: MutableMapping[str, Any],
    ) -> MutableMapping[str, Any]:
        """Extracts the message from the event dictionary and format it."""
        event = event_dict.get("event", "")
        try:
            event_data = json.loads(event)
            event_dict["event"] = event_data.get("event", event)
        except json.JSONDecodeError:
            ...
        return event_dict


Logger().configure_logger()


def get_logger(name: str | None = None) -> structlog.stdlib.BoundLogger:
    """Return a logger instance."""
    return structlog.get_logger(name)



from pathlib import Path

from readmeai.config.settings import ConfigLoader
from readmeai.generators.quickstart import QuickStartGenerator
from readmeai.ingestion.file_processor import FileProcessor
from readmeai.ingestion.metadata_extractor import MetadataExtractor
from readmeai.ingestion.models import RepositoryContext


class RepositoryProcessor:
    """
    Processes a repository to extract dependencies and metadata.
    """

    def __init__(self, config: ConfigLoader):
        self.config = config
        self.file_processor = FileProcessor(config)
        self.metadata_extractor = MetadataExtractor(config)
        self.quickstart_generator = QuickStartGenerator(config)

    async def process_repository(
        self, repo_path: Path | str | None = None
    ) -> RepositoryContext:
        """Process the repository and extract metadata."""
        repo_path = Path(str(repo_path))

        file_contexts = self.file_processor.process_files(repo_path)
        metadata = self.metadata_extractor.extract_metadata(file_contexts)
        language_counts = self.file_processor.count_languages(file_contexts)
        dependencies = self.file_processor.extract_dependencies(file_contexts)
        language_names = list(
            {file.language for file in file_contexts if file.language}
        )
        quickstart = self.quickstart_generator.generate(
            language_counts, metadata
        )
        dependencies_and_tools = (
            [tool for tool_group in metadata.values() for tool in tool_group]
            + language_names
            + dependencies
        )
        return RepositoryContext(
            files=file_contexts,
            dependencies=dependencies_and_tools,
            languages=language_names,
            language_counts=language_counts,
            metadata=metadata,
            quickstart=quickstart,
        )



"""Command-line interface entrypoint for the readme-ai package."""

import click

from readmeai.__main__ import readme_agent
from readmeai.cli import options
from readmeai.config.settings import ConfigLoader, GitSettings
from readmeai.logger import get_logger

config = ConfigLoader()
logger = get_logger(__name__)


@click.command()
@click.option(
    "--version",
    "-V",
    is_flag=True,
    callback=options.version_callback,
    expose_value=False,
    is_eager=True,
    help="Show the version and exit.",
)
@options.align
@options.api
@options.badge_color
@options.badge_style
@options.base_url
@options.context_window
@options.emojis
@options.header_style
@options.image
@options.model
@options.output
@options.rate_limit
@options.repository
@options.temperature
@options.toc_style
@options.top_p
@options.tree_depth
def main(
    align: str,
    api: str,
    badge_color: str,
    badge_style: str,
    base_url: str,
    context_window: int,
    emojis: bool,
    header_style: str,
    image: str,
    model: str,
    output: str,
    rate_limit: int,
    repository: str,
    temperature: float,
    toc_style: str,
    top_p: float,
    tree_depth: int,
) -> None:
    """Entry point for the readme-ai CLI application."""
    config.config.git = GitSettings(repository=repository)
    config.config.llm = config.config.llm.model_copy(
        update={
            "api": api,
            "base_url": base_url,
            "context_window": context_window,
            "model": model,
            "temperature": temperature,
            "top_p": top_p,
        },
    )
    config.config.md = config.config.md.model_copy(
        update={
            "align": align,
            "badge_color": badge_color,
            "badge_style": badge_style,
            "emojis": emojis,
            "header_style": header_style,
            "image": image,
            "toc_style": toc_style,
            "tree_depth": tree_depth,
        },
    )
    config.config.api.rate_limit = rate_limit

    logger.info(f"Pydantic settings: {config.__dict__.keys()}")
    logger.info(f"Repository settings: {config.config.git}")
    logger.info(f"LLM API settings: {config.config.llm}")

    readme_agent(config=config, output_file=output)


if __name__ == "__main__":
    main()



"""OpenAI API model handler implementation, with Ollama support."""

import os
from typing import Any

import aiohttp
import openai
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from readmeai.config.constants import LLMService
from readmeai.config.settings import ConfigLoader
from readmeai.ingestion.models import RepositoryContext
from readmeai.models.base import BaseModelHandler
from readmeai.models.tokens import token_handler


class OpenAIHandler(BaseModelHandler):
    """
    OpenAI API LLM implementation, with Ollama support.
    """

    def __init__(
        self, config_loader: ConfigLoader, context: RepositoryContext
    ) -> None:
        super().__init__(config_loader, context)
        self._model_settings()

    def _model_settings(self):
        self.host_name = self.config.llm.host_name
        self.localhost = self.config.llm.localhost
        self.model = self.config.llm.model
        self.path = self.config.llm.path
        self.tokens = self.config.llm.tokens
        self.top_p = self.config.llm.top_p

        if self.config.llm.api == LLMService.OPENAI.name:
            self.url = f"{self.host_name}{self.path}"
            self.client = openai.OpenAI(api_key=os.getenv("OPENAI_API_KEY"))

        elif self.config.llm.api == LLMService.OLLAMA.name:
            self.url = f"{self.localhost}{self.path}"
            self.client = openai.OpenAI(
                base_url=f"{self.localhost}v1",
                api_key=LLMService.OLLAMA.name,
            )

        self.headers = {"Authorization": f"Bearer {self.client.api_key}"}

    async def _build_payload(self, prompt: str, tokens: int) -> dict:
        """Build payload for POST request to OpenAI API."""
        return {
            "messages": [
                {
                    "role": "system",
                    "content": self.system_message,
                },
                {"role": "user", "content": prompt},
            ],
            "model": self.model,
            "max_tokens": tokens,
            "temperature": self.temperature,
        }

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_if_exception_type(
            (
                aiohttp.ClientError,
                aiohttp.ClientResponseError,
                aiohttp.ClientConnectorError,
                openai.OpenAIError,
            ),
        ),
    )
    async def _make_request(
        self,
        index: str | None,
        prompt: str | None,
        tokens: int | None,
        repo_files: Any,
    ) -> Any:
        """Processes OpenAI API LLM responses and returns generated text."""
        try:
            prompt = await token_handler(self.config, index, prompt, tokens)

            parameters = await self._build_payload(prompt, tokens)

            async with self._session.post(
                self.url,
                headers=self.headers,
                json=parameters,
            ) as response:
                response.raise_for_status()
                data = await response.json()
                data = data["choices"][0]["message"]["content"]
                self._logger.info(
                    f"Response from OpenAI for '{index}': {data}"
                )
                return index, data

        except (
            aiohttp.ClientError,
            aiohttp.ClientResponseError,
            aiohttp.ClientConnectorError,
            openai.OpenAIError,
        ) as e:
            self._logger.error(
                f"Error processing request for '{index}': {e!r}"
            )
            raise  # Re-raise for retry decorator

        except Exception as e:
            self._logger.error(f"Unexpected error for '{index}': {e!r}")
            return index, self.placeholder



"""Converts markdown syntax to HTML elements in a given piece of text.

Syntax supported:
- Bold (**text**)
- Italic (*text*)
- Links ([text](url))
- Headers (# Header)
- Lists (unordered and ordered)

Notes:
- Ensures compatibility with README-AI's HTML-based table content used in the README.md file.
- All regular expressions are precompiled at the module level to avoid repeated compilation.
"""

import re
from typing import Any

BOLD_RE: re.Pattern[str] = re.compile(r"\*\*(.+?)\*\*")

INLINE_CODE_RE: re.Pattern[str] = re.compile(r"`([^`\n]+)`")

ITALIC_RE: re.Pattern[str] = re.compile(r"\*(.+?)\*")

LINK_RE: re.Pattern[str] = re.compile(r"\[([^\]]+)\]\(([^)]+)\)")

ORDERED_LIST_RE: re.Pattern[str] = re.compile(
    r"^(\s*)\d+\.\s(.+)$", re.MULTILINE
)

UNORDERED_LIST_RE: re.Pattern[str] = re.compile(
    r"^(\s*)[-*+]\s(.+)$", re.MULTILINE
)

HEADER_RE: list[re.Pattern[str]] = [
    re.compile(f'^{"#" * i} (.+)$', re.MULTILINE) for i in range(6, 0, -1)
]


def markdown_to_html(text: str) -> str:
    """Convert any markdown syntax in the given text to HTML."""

    def process_inline(content: str) -> str:
        """Process inline markdown syntax."""

        def nested_format(match) -> str | Any:
            """Handle nested formatting within inline elements."""
            inner_content = process_inline(match.group(1))
            if match.re == BOLD_RE:
                return f"<strong>{inner_content}</strong>"
            elif match.re == ITALIC_RE:
                return f"<em>{inner_content}</em>"
            return match.group(0)

        content = INLINE_CODE_RE.sub(r"<code>\1</code>", content)
        content = BOLD_RE.sub(nested_format, content)
        content = ITALIC_RE.sub(nested_format, content)
        content = LINK_RE.sub(r'<a href="\2">\1</a>', content)
        return content

    for i, pattern in enumerate(HEADER_RE, start=1):
        text = pattern.sub(
            lambda m: f"<h{7-i}>{process_inline(m.group(1))}</h{7-i}>", text
        )

    lines = text.split("\n")
    in_list = False
    list_type = None
    list_indent = ""
    result = []

    for line in lines:
        unordered_match = UNORDERED_LIST_RE.match(line)
        ordered_match = ORDERED_LIST_RE.match(line)

        if unordered_match or ordered_match:
            match = unordered_match or ordered_match
            indent, content = match.groups()

            if not in_list:
                list_indent = indent
                list_type = "ul" if unordered_match else "ol"
                result.append(f"{indent}<{list_type}>")
                in_list = True
            elif (
                list_type != ("ul" if unordered_match else "ol")
                or indent != list_indent
            ):
                result.append(f"{list_indent}</{list_type}>")
                list_indent = indent
                list_type = "ul" if unordered_match else "ol"
                result.append(f"{indent}<{list_type}>")

            result.append(f"{indent}<li>{process_inline(content)}</li>")
        else:
            if in_list:
                result.append(f"{list_indent}</{list_type}>")
                in_list = False
                list_type = None
            result.append(process_inline(line))

    if in_list:
        result.append(f"{list_indent}</{list_type}>")

    return "\n".join(result)



#!/usr/bin/env python3

"""Orchestrates the pipeline for the README.md file generation."""

__package__ = "readmeai"

import asyncio
import contextlib
import tempfile
from collections.abc import Generator

from readmeai.config.constants import ImageOptions, LLMService
from readmeai.config.settings import ConfigLoader
from readmeai.errors import ReadmeGeneratorError
from readmeai.generators.builder import MarkdownBuilder
from readmeai.ingestion.models import RepositoryContext
from readmeai.ingestion.pipeline import RepositoryProcessor
from readmeai.logger import get_logger
from readmeai.models.dalle import DalleHandler
from readmeai.models.factory import ModelFactory
from readmeai.postprocessor import response_cleaner
from readmeai.readers.git.repository import load_data
from readmeai.utils.file_handler import FileHandler

_logger = get_logger(__name__)


@contextlib.contextmanager
def error_handler() -> Generator[None, None, None]:
    """Error handler wrapper for the README generation process."""
    try:
        yield
    except Exception as e:
        raise ReadmeGeneratorError(e) from e


def readme_agent(config: ConfigLoader, output_file: str) -> None:
    """Wrap asyncronous README generation process with context manager."""
    with error_handler():
        asyncio.run(readme_generator(config, output_file))


async def readme_generator(config: ConfigLoader, output_file: str) -> None:
    """Processes the repository and builds the README file."""
    with tempfile.TemporaryDirectory() as temp_dir:
        repo_path = await load_data(config.config.git.repository, temp_dir)

        processor: RepositoryProcessor = RepositoryProcessor(config=config)
        context: RepositoryContext = await processor.process_repository(
            repo_path=repo_path
        )

        log_repository_context(context)

        async with ModelFactory.get_backend(config, context).use_api() as llm:
            responses = await llm.batch_request()
            (
                file_summaries,
                features,
                overview,
                slogan,
            ) = responses
            config.config.md.features = config.config.md.features.format(
                response_cleaner.format_markdown_table(features),
            )
            config.config.md.overview = config.config.md.overview.format(
                response_cleaner.process_markdown(overview),
            )
            config.config.md.slogan = response_cleaner.remove_quotes(slogan)

        if should_generate_image(config):
            await generate_image(config)
        if config.config.md.image in [None, "", ImageOptions.LLM.value]:
            config.config.md.image = ImageOptions.BLUE.value

        readme_md_content = MarkdownBuilder(
            config, context, file_summaries, temp_dir
        ).build()

        FileHandler().write(output_file, readme_md_content)

        log_process_completion(output_file)


async def generate_image(config: ConfigLoader) -> None:
    """Generates an image using the DALL-E model."""
    async with DalleHandler(config) as dalle:
        image_url = await dalle._make_request()
        config.config.md.image = await dalle.download(image_url)


def should_generate_image(config: ConfigLoader) -> bool:
    """Determines if the user enabled LLM image generation."""
    return (
        config.config.md.image == ImageOptions.LLM.value
        and config.config.llm.api != LLMService.OFFLINE.value
    )


def log_repository_context(context: RepositoryContext) -> None:
    """Logs a snippet of the processed repository context data."""
    _logger.info(f"Total files analyzed: {len(context.files)}")
    _logger.info(f"Metadata extracted: {context.metadata}")
    _logger.info(f"Dependencies: {context.dependencies}")
    _logger.info(f"Languages: {context.language_counts}")


def log_process_completion(output_file: str) -> None:
    """Logs the completion of the README generation process."""
    _logger.info("README.md file generated successfully.")
    _logger.info(f"Output file saved @ {output_file}")
    _logger.info("Share with us @ github.com/eli64s/readme-ai/discussions")



from pathlib import Path

from readmeai.logger import get_logger

_logger = get_logger(__name__)


def build_submodule_disclosure_widget(
    data: dict[str, dict | list[tuple[str, str]]],
    repo_path: str | Path,
    repo_url: str,
) -> str:
    """
    Builds expandable sections via <details> HTML element,
    for each module with nested submodules using HTML tables.
    """
    if not data:
        _logger.warning("No file summaries found.")
        return ""

    is_local_repo = Path(repo_path).exists()

    project_name = (
        Path(repo_path).name
        if is_local_repo
        else repo_url.rstrip("/").split("/")[-1]
    )

    sections = [
        "<details open>",
        f"\t<summary><b><code>{project_name.upper()}/</code></b></summary>",
    ]

    for module_name, module_data in data.items():
        # Root module
        if module_name == {project_name}:
            section = [
                f"\t<details> <!-- {module_name} Submodule -->",
                f"\t\t<summary><b>{module_name}</b></summary>",
                "\t\t<blockquote>",
                "\t\t\t<table>",
            ]
            # Generate table rows for root files
            section.extend(
                _generate_table_rows(
                    module_data,
                    repo_path,
                    is_local_repo,
                    repo_url,
                    indent="\t\t\t",
                )
            )
            section.extend(
                ("\t\t\t</table>", "\t\t</blockquote>", "\t</details>")
            )
        else:
            # Handle other modules
            section = [
                f"\t<details> <!-- {module_name} Submodule -->",
                f"\t\t<summary><b>{module_name}</b></summary>",
                "\t\t<blockquote>",
            ]
            # Generate content for submodules or files
            section.extend(
                _generate_nested_module_content(
                    module_data,
                    repo_path,
                    is_local_repo,
                    repo_url,
                    indent="\t\t\t",
                )
            )
            section.extend(("\t\t</blockquote>", "\t</details>"))

        sections.extend(section)

    sections.append("</details>\n")

    return "\n".join(sections)


def format_code_summaries(
    placeholder: str,
    code_summaries: list[tuple[str, str]],
) -> list:
    """Converts the given code summaries into a formatted list."""
    return [
        (module, summary_text)
        if isinstance(summary, tuple) and len(summary) == 2
        else (summary, placeholder)
        for summary in code_summaries
        for module, summary_text in (
            [summary]
            if isinstance(summary, tuple) and len(summary) == 2
            else [(summary, placeholder)]
        )
    ]


def format_summary(summary: str) -> str:
    """
    Formats the summary with multi-line support if needed.
    """
    lines = summary.strip().split(". ")
    if len(lines) > 1:
        return "<br>".join(f"- {line.strip()}" for line in lines)
    return summary.strip()


def _generate_nested_module_content(
    module_data: dict[str, dict | list[tuple[str, str]]]
    | list[tuple[str, str]],
    repo_path: str | Path,
    is_local_repo: bool,
    repo_url: str,
    indent: str = "",
) -> list[str]:
    """Generates nested content for modules and submodules using HTML tables."""
    content = []

    if isinstance(module_data, list):
        # module_data is a list of files
        content.append(f"{indent}<table>")
        content.extend(
            _generate_table_rows(
                module_data,
                repo_path,
                is_local_repo,
                repo_url,
                indent=indent,
            )
        )
        content.append(f"{indent}</table>")
    elif isinstance(module_data, dict):
        # Check if there are files at this module level
        files = module_data.get("", [])
        if files:
            content.append(f"{indent}<table>")
            content.extend(
                _generate_table_rows(
                    files,
                    repo_path,
                    is_local_repo,
                    repo_url,
                    indent=indent,
                )
            )
            content.append(f"{indent}</table>")
        # Process submodules
        for submodule_name, submodule_data in module_data.items():
            if submodule_name == "":
                continue
            content.extend(
                (
                    f"{indent}<details>",
                    f"{indent}\t<summary><b>{submodule_name}</b></summary>",
                    f"{indent}\t<blockquote>",
                )
            )
            # Recurse into submodule
            content.extend(
                _generate_nested_module_content(
                    submodule_data,
                    repo_path,
                    is_local_repo,
                    repo_url,
                    indent=indent + "\t\t",
                )
            )
            content.extend((f"{indent}\t</blockquote>", f"{indent}</details>"))
    else:
        _logger.warning(
            f"Unexpected data type in module data: {type(module_data)}"
        )
    return content


def generate_nested_module_tables(
    summaries: list[tuple[str, str]],
    project_path: str | Path,
    repository_url: str,
) -> str:
    """Create structured Markdown tables with nested submodules."""
    summaries_by_module = group_summaries_by_nested_module(summaries)
    return build_submodule_disclosure_widget(
        summaries_by_module, project_path, repository_url
    )


def _generate_table_rows(
    files: list[tuple[str, str]],
    repo_path: str | Path,
    is_local_repo: bool,
    repo_url: str,
    indent: str = "",
) -> list[str]:
    """Generates table rows for files."""
    content = []
    for file, summary in files:
        file_name = Path(file).name
        file_link = _get_file_hyperlink(
            file, repo_path, is_local_repo, repo_url
        )
        formatted_summary = format_summary(summary)
        content.append(
            f"{indent}<tr>"
            f"\n{indent}\t<td><b><a href='{file_link}'>{file_name}</a></b></td>"
            f"\n{indent}\t<td>{formatted_summary}</td>"
            f"\n{indent}</tr>"
        )
    return content


def _get_file_hyperlink(
    file_path_str: str,
    repo_path: str | Path,
    is_local: bool,
    repo_url: str,
) -> str:
    """Generates a hyperlink to the file in the remote git repository."""
    if not is_local:
        return f"{repo_url.rstrip('/')}/blob/master/{file_path_str}"
    file_path = Path(repo_path) / file_path_str
    return f"{file_path.as_posix()}"


def group_summaries_by_nested_module(
    summaries: list[tuple[str, str]],
) -> dict[str, dict | list[tuple[str, str]]]:
    """Group code summaries by their nested module structure."""
    module_map: dict[str, dict | list[tuple[str, str]]] = {"__root__": []}

    for module, summary in summaries:
        if not isinstance(module, str) or not isinstance(summary, str):
            _logger.error(f"Invalid entry: ({module}, {summary}). Skipping...")
            continue

        parts = Path(module).parts

        if len(parts) == 1:
            # File in the root directory
            module_map["__root__"].append((module, summary))
        else:
            current = module_map
            for part in parts[:-1]:
                current = current.setdefault(part, {})
            current.setdefault("", []).append((module, summary))

    return module_map



import asyncio
import os
import platform
import shutil
from pathlib import Path


async def remove_directory(path: Path) -> None:
    """Remove a temporary directory and its contents."""
    if platform.system() == "Windows":
        os.system(f'rmdir /S /Q "{path}"')
    else:
        await asyncio.to_thread(shutil.rmtree, path, ignore_errors=True)


async def remove_hidden_contents(directory: Path) -> None:
    """Remove hidden files and directories from a specified directory."""
    for item in directory.iterdir():
        if ".github" in item.parts:
            continue
        if item.name.startswith("."):
            if item.is_dir():
                shutil.rmtree(item)
            else:
                item.unlink()



"""
Parser for Rust cargo.toml dependency files.
"""

import sys

from readmeai.parsers.base import BaseFileParser

if sys.version_info < (3, 11):
    import tomli as toml
else:
    import tomllib as toml


class CargoTomlParser(BaseFileParser):
    """
    Parser for Rust cargo.toml dependency files.
    """

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> list[str]:
        """Extract packages names from Rust TOML files."""
        try:
            data = toml.loads(content)
            dependencies = []

            if "dependencies" in data:
                dependencies.extend(data["dependencies"].keys())
            if "dev-dependencies" in data:
                dependencies.extend(data["dev-dependencies"].keys())

            for key in data:
                if key.startswith("dependencies.") and isinstance(
                    data[key],
                    dict,
                ):
                    dependencies.extend(data[key].keys())

            return dependencies

        except Exception as exc:
            return self.handle_parsing_error(f"cargo.toml: {exc!s}")



import asyncio
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, Generator
from contextlib import asynccontextmanager
from typing import Any

import aiohttp

from readmeai.config.constants import LLMService
from readmeai.config.settings import ConfigLoader
from readmeai.ingestion.models import RepositoryContext
from readmeai.logger import get_logger
from readmeai.models.prompts import (
    get_prompt_context,
    set_additional_contexts,
    set_summary_context,
)
from readmeai.models.tokens import update_max_tokens


class BaseModelHandler(ABC):
    """
    Interface for LLM API handler implementations.
    """

    def __init__(
        self, config_loader: ConfigLoader, context: RepositoryContext
    ) -> None:
        self._logger = get_logger(__name__)
        self._session: aiohttp.ClientSession | None = None
        self.config = config_loader.config
        self.placeholder = self.config.md.placeholder
        self.prompts = config_loader.prompts
        self.max_tokens = self.config.llm.tokens
        self.rate_limit = self.config.api.rate_limit
        self.rate_limit_semaphore = asyncio.Semaphore(self.rate_limit)
        self.temperature = self.config.llm.temperature
        self.system_message = self.config.api.system_message
        self.repo_context = context
        self.dependencies = context.dependencies
        self.documents = [
            (file.path, file.content)
            for file in context.files
            if ".lock" not in file.name
        ]

    @asynccontextmanager
    async def use_api(self) -> AsyncGenerator[Any, None]:
        """Async context manager for managing HTTP client lifecycle."""
        async with aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=None),
        ) as session:
            self._session = session
            try:
                yield self
            finally:
                await self.close()

    async def close(self) -> None:
        """Closes the HTTP client session."""
        if self._session:
            await self._session.close()
            self._logger.debug("HTTP client closed.")
        else:
            self._logger.debug("HTTP client is already closed.")

    @abstractmethod
    async def _model_settings(self) -> None:
        """Initializes LLM API settings for a given service."""
        ...

    @abstractmethod
    async def _build_payload(self, prompt: str, tokens: int) -> dict[str, Any]:
        """Builds the payload for the POST request to the LLM API."""
        ...

    @abstractmethod
    async def _make_request(
        self,
        index: str | None,
        prompt: str | None,
        tokens: int | None,
        repo_files: list[tuple[str, str]] | None,
    ) -> Any:
        """Handles LLM API response and returns the generated text."""
        ...

    async def batch_request(self) -> list[tuple[str, str]]:
        """Generates a batch of prompts and processes the responses."""
        if self.config.llm.api == LLMService.OFFLINE.name:
            return await self._make_request(
                index=None,
                prompt=None,
                tokens=None,
                repo_files=self.documents,
            )

        summaries_prompts = await set_summary_context(
            self.config,
            self.documents,
        )
        summaries_responses = await self._batch_prompts(summaries_prompts)

        additional_prompts = await set_additional_contexts(
            self.config,
            self.repo_context,
            summaries_responses,
        )
        additional_responses = await self._batch_prompts(additional_prompts)

        return summaries_responses + additional_responses

    async def _batch_prompts(
        self,
        prompts: Any,
        batch_size: int = 10,
    ) -> list[tuple[str, str]]:
        """Processes a batch of prompts and returns the generated text."""
        responses = []

        for batch in self._generate_batches(prompts, batch_size):
            batch_responses = await asyncio.gather(
                *[self._process_batch(prompt) for prompt in batch],
                return_exceptions=False,
            )
            responses.extend(batch_responses)

        return responses

    def _generate_batches(
        self,
        items: list[Any],
        batch_size: int,
    ) -> Generator[list[Any], None, None]:
        """Generates batches of items to be processed."""
        for i in range(0, len(items), batch_size):
            yield items[i : i + batch_size]

    async def _process_batch(self, prompt: dict[str, Any]) -> Any:
        """Processes a single prompt and returns the generated text."""
        if prompt["type"] == "file_summary":
            return await self._make_request_code_summary(
                prompt["context"],
            )
        else:
            formatted_prompt = get_prompt_context(
                self.prompts,
                prompt["type"],
                prompt["context"],
            )
            tokens = update_max_tokens(
                self.config.llm.tokens,
                formatted_prompt,
            )
            _, summary = await self._make_request(
                prompt["type"],
                formatted_prompt,
                tokens,
                None,
            )
            return summary

    async def _make_request_code_summary(
        self,
        file_context: list[tuple[str, str]],
    ) -> tuple:
        """Generates code summaries for each file in the project."""
        file_summaries = []

        for file_path, file_content in file_context["repo_files"]:
            prompt = self.prompts["prompts"]["file_summary"].format(
                self.config.md.tree,
                file_path,
                file_content,
            )
            tokens = update_max_tokens(self.config.llm.tokens, prompt)
            _, summary_or_error = await self._make_request(
                file_path,
                prompt,
                tokens,
                None,
            )
            file_summaries.append((file_path, summary_or_error))

        return file_summaries



"""
Enum classes that store information settings for the LLM
API service providers, badge styles, and image options.
"""

import enum


class BadgeStyleOptions(str, enum.Enum):
    """
    Badge icon styles for the project README.
    """

    DEFAULT = "default"
    FLAT = "flat"
    FLAT_SQUARE = "flat-square"
    FOR_THE_BADGE = "for-the-badge"
    PLASTIC = "plastic"
    SKILLS = "skills"
    SKILLS_LIGHT = "skills-light"
    SOCIAL = "social"


class HeaderStyleOptions(str, enum.Enum):
    """
    Enum of supported 'Header' template styles for the README file.
    """

    ASCII = "ascii"
    ASCII_BOX = "ascii_box"
    CLASSIC = "classic"
    COMPACT = "compact"
    MODERN = "modern"
    SVG = "svg"


class ImageOptions(str, enum.Enum):
    """
    Default image options for the project logo.
    """

    CUSTOM = "CUSTOM"
    LLM = "LLM"
    BLACK = "https://img.icons8.com/external-tal-revivo-regular-tal-revivo/96/external-readme-is-a-easy-to-build-a-developer-hub-that-adapts-to-the-user-logo-regular-tal-revivo.png"
    BLUE = "https://raw.githubusercontent.com/PKief/vscode-material-icon-theme/ec559a9f6bfd399b82bb44393651661b08aaf7ba/icons/folder-markdown-open.svg"
    CLOUD = "https://cdn-icons-png.flaticon.com/512/6295/6295417.png"
    GRADIENT = "https://img.icons8.com/?size=512&id=55494&format=png"
    GREY = "https://img.icons8.com/external-tal-revivo-filled-tal-revivo/96/external-markdown-a-lightweight-markup-language-with-plain-text-formatting-syntax-logo-filled-tal-revivo.png"
    PURPLE = "https://img.icons8.com/external-tal-revivo-duo-tal-revivo/100/external-markdown-a-lightweight-markup-language-with-plain-text-formatting-syntax-logo-duo-tal-revivo.png"


class TocStyleOptions(str, enum.Enum):
    """
    Enum of supported 'Table of Contents' templates for the README file.
    """

    BULLET = "bullet"
    FOLD = "fold"
    LINKS = "links"
    NUMBER = "number"
    ROMAN = "roman"


class LLMService(str, enum.Enum):
    """
    LLM API service providers.
    """

    ANTHROPIC = "ANTHROPIC"
    GEMINI = "GEMINI"
    OLLAMA = "OLLAMA"
    OPENAI = "OPENAI"
    OFFLINE = "OFFLINE"


class ServiceAuthKeys(str, enum.Enum):
    """
    Environment variable names associated with a LLM API key.
    """

    ANTHROPIC_API_KEY = "ANTHROPIC_API_KEY"
    GOOGLE_API_KEY = "GOOGLE_API_KEY"
    OLLAMA_HOST = "OLLAMA_HOST"
    OPENAI_API_KEY = "OPENAI_API_KEY"



"""The 'Offline Mode' model handler runs the CLI without LLM API connection."""

from typing import Any

from readmeai.config.settings import ConfigLoader
from readmeai.ingestion.models import RepositoryContext
from readmeai.models.base import BaseModelHandler


class OfflineHandler(BaseModelHandler):
    """
    OfflineMode model handler implementation.
    """

    def __init__(
        self, config_loader: ConfigLoader, context: RepositoryContext
    ) -> None:
        super().__init__(config_loader, context)

    async def _model_settings(self) -> None: ...

    async def _build_payload(self, prompt: str, tokens: int) -> dict[str, Any]:
        """Builds the payload for the POST request to the LLM API."""
        return {}

    async def _make_request(
        self,
        index: str | None,
        prompt: str | None,
        tokens: int | None,
        repo_files: Any,
    ) -> Any:
        """Inserts placeholder text where the LLM API response would be."""
        files = [(file_path, self.placeholder) for file_path, _ in repo_files]
        return (files, self.placeholder, self.placeholder, self.placeholder)



"""
Dependency file parsers for Swift projects.
"""

import re

from readmeai.parsers.base import BaseFileParser


class SwiftPackageParser(BaseFileParser):
    """
    Parser for Swift Package.swift files.
    """

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> list[str]:
        """Extracts package names from a Package.swift file."""
        try:
            package_names = set()
            lines = content.splitlines()

            for line in lines:
                line = line.strip()

                if ".package(url:" in line or ".package(name:" in line:
                    url_match = re.search(r'url:\s*"([^"]+)"', line)
                    name_match = re.search(r'name:\s*"([^"]+)"', line)

                    if url_match:
                        url = url_match.group(1)
                        package_name = self.extract_package_name_from_url(url)
                        package_names.add(package_name)
                    elif name_match:
                        package_name = name_match.group(1)
                        package_names.add(package_name)

                elif "dependencies:" in line:
                    dep_match = re.findall(r"\"([^\"]+)\"", line)
                    package_names.update(dep_match)

            return list(package_names)

        except Exception as exc:
            return self.handle_parsing_error(f"Package.swift: {exc!s}")

    @staticmethod
    def extract_package_name_from_url(url: str) -> str:
        """Extracts the package name from a GitHub URL."""
        package_name = url.split("/")[-1]
        if package_name.endswith(".git"):
            package_name = package_name[:-4]
        return package_name



from typing import ClassVar

from .base import BaseFileParser, DefaultParser
from .cpp import CMakeParser, ConfigureAcParser, MakefileAmParser
from .docker import DockerComposeParser, DockerfileParser
from .go import GoModParser
from .gradle import BuildGradleKtsParser, BuildGradleParser
from .maven import MavenParser
from .npm import PackageJsonParser
from .properties import PropertiesParser
from .python import RequirementsParser, TomlParser, YamlParser
from .rust import CargoTomlParser
from .swift import SwiftPackageParser

ParserRegistryType = dict[str, type[BaseFileParser]]


class ParserFactory:
    """
    Factory for creating dependency file parser callable objects.
    """

    _parsers: ClassVar[dict[str, type[BaseFileParser]]] = {
        # Python
        "Pipfile": TomlParser,
        "pyproject.toml": TomlParser,
        "requirements.in": RequirementsParser,
        "requirements.txt": RequirementsParser,
        "requirements-dev.txt": RequirementsParser,
        "requirements-test.txt": RequirementsParser,
        "requirements-prod.txt": RequirementsParser,
        "dev-requirements.txt": RequirementsParser,
        "environment.yml": YamlParser,
        "environment.yaml": YamlParser,
        "poetry.lock": DefaultParser,
        "pdm.lock": DefaultParser,
        # C/C++
        "CMakeLists.txt": CMakeParser,
        "configure.ac": ConfigureAcParser,
        "Makefile.am": MakefileAmParser,
        # JavaScript/Node.js
        "package.json": PackageJsonParser,
        # Kotlin/Kotlin DSL
        "build.gradle": BuildGradleParser,
        "build.gradle.kts": BuildGradleKtsParser,
        # Go
        "go.mod": GoModParser,
        # Java
        "pom.xml": MavenParser,
        # Rust
        "Cargo.toml": CargoTomlParser,
        # Swift
        "Package.swift": SwiftPackageParser,
        # Docker
        "Dockerfile": DockerfileParser,
        "docker-compose.yaml": DockerComposeParser,
        "docker-compose.yml": DockerComposeParser,
        # Properties
        ".properties": PropertiesParser,
    }

    @classmethod
    def register_parser(
        cls, file_name: str, parser_class: type[BaseFileParser]
    ):
        """Register a parser for the given file name."""
        cls._parsers[file_name] = parser_class

    @classmethod
    def create_parser(cls, file_name: str) -> BaseFileParser:
        """Create a parser for the given file name."""
        if parser_class := cls._parsers.get(file_name):
            return parser_class()
        else:
            return DefaultParser()



"""Utility methods to build prompts for LLM text generation."""

from readmeai.config.settings import Settings
from readmeai.ingestion.models import RepositoryContext
from readmeai.logger import get_logger

_logger = get_logger(__name__)


def get_prompt_context(prompts: dict, prompt_type: str, context: dict) -> str:
    """Generates a prompt for the LLM API."""
    prompt_template = get_prompt_template(prompts, prompt_type)

    if not prompt_template:
        _logger.error(f"Prompt type '{prompt_type}' not found.")
        return ""

    return inject_prompt_context(prompt_template, context)


def get_prompt_template(prompts: dict, prompt_type: str) -> str:
    """Retrieves the template for the given prompt type."""
    prompt_templates = {
        "features_table": prompts["prompts"]["features_table"],
        "overview": prompts["prompts"]["overview"],
        "slogan": prompts["prompts"]["slogan"],
    }
    return prompt_templates.get(prompt_type, "")


def inject_prompt_context(template: str, context: dict) -> str:
    """Formats the template with the provided context."""
    try:
        return template.format(*[context[key] for key in context])
    except KeyError as exc:
        _logger.error(f"Missing context for prompt key: {exc}")
        return ""


async def set_additional_contexts(
    config: Settings,
    repo_context: RepositoryContext,
    file_summaries: list[tuple[str, str]],
) -> list[dict]:
    """Generates additional prompts (features, overview, slogan) for LLM."""
    return [
        {"type": prompt_type, "context": context}
        for prompt_type, context in [
            (
                "features_table",
                {
                    "name": config.git.name,
                    "dependencies": repo_context.dependencies,
                    "quickstart": repo_context.quickstart,
                    "file_summary": file_summaries,
                },
            ),
            (
                "overview",
                {
                    "name": config.git.name,
                    "file_summary": file_summaries,
                },
            ),
            (
                "slogan",
                {
                    "name": config.git.name,
                    "repo": config.git.repository,
                    "file_summary": file_summaries,
                },
            ),
        ]
    ]


async def set_summary_context(
    config: Settings,
    repo_files: list[tuple[str, str]],
) -> list[dict]:
    """Generates the summary prompts to be used by the LLM API."""
    return [
        {"type": prompt_type, "context": context}
        for prompt_type, context in [
            (
                "file_summary",
                {
                    "tree": config.md.tree,
                    "repo_files": repo_files,
                },
            ),
        ]
    ]



"""Python dependency file parsers methods."""

import re
import sys

import yaml

from readmeai.parsers.base import BaseFileParser

if sys.version_info < (3, 11):
    import tomli as toml
else:
    import tomllib as toml


class RequirementsParser(BaseFileParser):
    """
    Parser for requirements.txt files.
    """

    def __init__(self) -> None:
        super().__init__()

    def parse(self, content: str) -> list[str]:
        """
        Extracts package names from requirements.txt file.
        Excludes the version specifiers and comments.
        """
        try:
            lines = content.splitlines()
            package_names = []
            for line in lines:
                line = line.split("#", 1)[0].strip()
                if not line:
                    continue
                if match := re.match(r"([a-zA-Z0-9\-_]+)", line):
                    package_names.append(match[1])
            return package_names

        except re.error as exc:
            return self.handle_parsing_error(f"requirements.txt: {exc!s}")


class TomlParser(BaseFileParser):
    """
    Parser for Python TOML dependency files.

    - Handles build systems:
        - Pipenv, Poetry, Flit, Hatch, etc.
    """

    def parse(self, content: str) -> list[str]:
        """Extracts all package dependencies from a TOML file."""
        try:
            data = toml.loads(content)
            dependencies = []

            # Look for dependencies in different possible sections
            # Pipfile-style
            if "packages" in data:
                dependencies.extend(data.get("packages", {}).keys())
            if "dev-packages" in data:
                dependencies.extend(data.get("dev-packages", {}).keys())

            # pyproject.toml-style (Poetry, Hatch, Flit)
            if "project" in data:
                dependencies.extend(
                    self.extract_package_names(
                        data["project"].get("dependencies", [])
                    )
                )
                if "optional-dependencies" in data["project"]:
                    for group_deps in data["project"][
                        "optional-dependencies"
                    ].values():
                        dependencies.extend(
                            self.extract_package_names(group_deps)
                        )

            # For build-system specific tools like Poetry
            if "tool" in data:
                if "poetry" in data["tool"]:
                    poetry_data = data["tool"]["poetry"]
                    dependencies.extend(
                        poetry_data.get("dependencies", {}).keys()
                    )

                    # Add dev-dependencies and any group dependencies
                    for group in ["dev-dependencies", "group"]:
                        if group in poetry_data:
                            group_data = poetry_data[group]
                            if isinstance(group_data, dict):
                                for section in group_data.values():
                                    dependencies.extend(
                                        section.get("dependencies", {}).keys()
                                    )

                if "hatch" in data["tool"]:
                    hatch_data = data["tool"]["hatch"].get("envs", {})
                    for env, env_data in hatch_data.items():
                        dependencies.extend(env_data.get("dependencies", []))
                        if "extra-dependencies" in env_data:
                            dependencies.extend(env_data["extra-dependencies"])

            # For Cargo.toml (Rust) style
            if "dependencies" in data:
                dependencies.extend(data.get("dependencies", {}).keys())
            if "dev-dependencies" in data:
                dependencies.extend(data.get("dev-dependencies", {}).keys())

            return list(set(dependencies))

        except Exception as exc:
            return self.handle_parsing_error(exc)

    def extract_package_names(self, dependencies: list[str]) -> list[str]:
        """Helper method to extract package names from a list of dependencies."""
        package_names = []
        for dep in dependencies:
            # Handle cases where dependencies might be in the format 'package_name>=1.0.0'
            if isinstance(dep, str):
                package_name = (
                    dep.split(">=")[0].split("<=")[0].split("==")[0].strip()
                )
                package_names.append(package_name)
            elif isinstance(dep, dict):
                package_names.extend(dep.keys())
        return package_names

    def extract_package_names(self, dependencies: list[str]) -> list[str]:
        """Helper method to extract package names from dependency strings."""
        package_names = []
        for dep in dependencies:
            match = re.match(r"([a-zA-Z0-9\-_]+)", dep.split(";")[0].strip())
            if match:
                package_names.append(match.group(1))
        return package_names


class YamlParser(BaseFileParser):
    """
    Parser for Python YAML based dependency files i.e. environment.yml
    """

    def parse(self, content: str) -> list[str]:
        """Extracts package names from environment.yml file."""
        try:
            data = yaml.safe_load(content)

            if isinstance(data, dict) and "dependencies" in data:
                dependencies = []

                for package in data["dependencies"]:
                    if isinstance(package, str):
                        dependencies.append(
                            package.split("=")[0].split(">")[0].split("<")[0],
                        )
                    elif isinstance(package, dict):
                        for pip_package in package.values():
                            if isinstance(pip_package, list):
                                for pip_dep in pip_package:
                                    dependencies.append(pip_dep.split("==")[0])

            return dependencies

        except yaml.YAMLError as exc:
            return self.handle_parsing_error(exc)



"""Filter files based on a default ignore list."""

from pathlib import Path


def is_excluded(ignore_list: dict, file_path: Path, repo_path: Path) -> bool:
    """Check if the file should be ignored based on the ignore list."""
    relative_path = file_path.relative_to(repo_path)

    for ignored_dir in ignore_list.get("directories", []):
        if ignored_dir in relative_path.parts:
            return True

    if file_path.suffix.lstrip(".") in ignore_list.get("extensions", []):
        return True

    return file_path.name in ignore_list.get("files", [])



"""Utilities to remove emojis from the default markdown template headers.

Example:
    Default Header:
            ## 🚀 Getting Started
    After Removing Emojis:
            ## Getting Started

Feature is enabled by default and can be disabled
in the CLI by passing '--emoji'. or '-e' flag.
"""

import re

EMOJI_PATTERN = re.compile(
    pattern="["
    "\U0001f600-\U0001f64f"  # emoticons
    "\U0001f300-\U0001f5ff"  # symbols & pictographs
    "\U0001f680-\U0001f6ff"  # transport & map symbols
    "\U0001f700-\U0001f77f"  # alchemical symbols
    "\U0001f780-\U0001f7ff"  # Geometric Shapes Extended
    "\U0001f800-\U0001f8ff"  # Supplemental Arrows-C
    "\U0001f900-\U0001f9ff"  # Supplemental Symbols and Pictographs
    "\U0001fa00-\U0001fa6f"  # Chess Symbols
    "\U0001fa70-\U0001faff"  # Symbols and Pictographs Extended-A
    "\U00002702-\U000027b0"  # Dingbats
    "\U000024c2-\U0001f251"  # flags (iOS)
    "]+",
    flags=re.UNICODE,
)


def remove_emojis(md_content: list[str]) -> list[str]:
    """Removes emojis from the content list."""
    modified_content = []

    for section in md_content:
        lines = section.split("\n")
        for index, line in enumerate(lines):
            if (
                line.startswith("#")
                or "Table of Contents" in section
                or "Quick Links" in section
            ):
                lines[index] = EMOJI_PATTERN.sub("", line)
        modified_content.append("\n".join(lines))

    return modified_content



from __future__ import annotations


class ReadmeAIError(Exception):
    """
    Base class for exceptions in this module.
    """

    ...


class ReadmeGeneratorError(Exception):
    """
    Raised when an error occurs during README generation.
    """

    def __init__(self, message, *args):
        self.message = f"Error generating README: {message}"
        super().__init__(self.message)


# ----------------- CLI ----------------------------------


class CLIError(ReadmeAIError):
    """Exceptions related to the CLI."""

    def __init__(self, message, *args):
        super().__init__(f"Invalid option provided to CLI: {message}", *args)


# ----------------- File System ----------------------------------


class FileSystemError(ReadmeAIError):
    """
    Exceptions related to file system operations.
    """

    def __init__(self, message, *args):
        super().__init__(f"File system error: {message}", *args)


class FileReadError(FileSystemError):
    """
    Raised when a file cannot be read.
    """

    ...


class FileWriteError(FileSystemError):
    """
    Raised when a file cannot be written to.
    """

    ...


# ----------------- Git ----------------------------------


class GitValidationError(ReadmeAIError):
    """
    Base class errors validating Git repositories.
    """

    ...


class GitCloneError(GitValidationError):
    """
    Raised when a Git repository cannot be cloned.
    """

    def __init__(self, repository: str, *args):
        self.repository = repository
        super().__init__(f"Failed to clone repository: {repository}", *args)


class GitURLError(GitValidationError):
    """
    Raised when an invalid Git repository URL is provided.
    """

    def __init__(self, url: str, *args):
        self.url = url
        super().__init__(f"Invalid Git repository URL: {url}", *args)


class InvalidRepositoryError(GitValidationError):
    """
    Raised when an invalid repository is provided.
    """

    def __init__(self, repository: str, *args):
        self.repository = repository
        super().__init__(f"Invalid repository provided: {repository}", *args)


class UnsupportedGitHostError(GitValidationError):
    """
    Raised when an unsupported Git host is provided.
    """

    def __init__(self, host: str, *args):
        self.host = host
        super().__init__(f"Unsupported Git host: {host}", *args)


# ----------------- Repository ----------------------------------


class RepositoryProcessingError(ReadmeAIError):
    """
    Raised when an error occurs during repository processing.
    """

    ...


# ----------------- LLM API ----------------------------------


class UnsupportedServiceError(ReadmeAIError):
    """
    Raised when an unsupported LLM service is provided.
    """

    def __init__(self, message, *args):
        super().__init__(message, *args)



"""File handler module to read and write various file formats."""

import functools
import json
from collections.abc import Callable
from pathlib import Path
from typing import Any

import yaml

from readmeai.errors import FileReadError, FileWriteError
from readmeai.utils.helpers import is_available


class FileHandler:
    """
    File I/O support for md, json, toml, txt, and yaml formats.
    """

    def __init__(self):
        self.file_actions = {
            "html": {"read": self.read_text, "write": self.write_text},
            "json": {"read": self.read_json, "write": self.write_json},
            "md": {"read": self.read_markdown, "write": self.write_markdown},
            "toml": {"read": self.read_toml, "write": self.write_toml},
            "txt": {"read": self.read_text, "write": self.write_text},
            "yaml": {"read": self.read_yaml, "write": self.write_yaml},
        }
        self.cache = {}
        self.read_json = functools.lru_cache(maxsize=100)(self.read_json)
        self.read_toml = functools.lru_cache(maxsize=100)(self.read_toml)

    def read(self, file_path: str | Path) -> Any:
        """Read the content of a file."""
        if file_path in self.cache:
            return self.cache[file_path]

        try:
            file_extension = str(file_path).rsplit(".", maxsplit=1)[-1]
            reader = self.get_action(file_extension, "read")
            content = reader(file_path)
            self.cache[file_path] = content
            return content
        except Exception as exc:
            raise FileReadError(exc, file_path) from exc

    def write(self, file_path: str | Path, content: Any) -> None:
        """Write the content to a file."""
        try:
            file_extension = str(file_path).rsplit(".", maxsplit=1)[-1]
            writer = self.get_action(file_extension, "write")
            writer(file_path, content)
        except Exception as exc:
            raise FileWriteError(exc, file_path) from exc

    def get_action(
        self, file_extension: str, action_type: str
    ) -> Callable[[str], Any]:
        """Get the method for the passed file extension and I/O operation."""
        file_actions = self.file_actions.get(file_extension)
        if not file_actions:
            raise ValueError(f"Unsupported file type: {file_extension}")

        action = file_actions.get(action_type)
        if not action:
            raise ValueError(f"Unsupported action type: {action_type}")

        return action

    @staticmethod
    def read_html(file_path: str | Path) -> str:
        """Read the content of an HTML file."""
        with open(file_path, encoding="utf-8") as file:
            return file.read()

    @staticmethod
    @functools.lru_cache(maxsize=100)
    def read_json(file_path: str | Path) -> dict[str, Any]:
        """Read the content of a JSON file."""
        with open(file_path, encoding="utf-8") as file:
            return json.load(file)

    @staticmethod
    def read_markdown(file_path: str | Path) -> str:
        """Read the content of a Markdown file."""
        with open(file_path, encoding="utf-8") as file:
            return file.read()

    @staticmethod
    @functools.lru_cache(maxsize=100)
    def read_toml(file_path: str | Path) -> dict[str, Any]:
        """Read the content of a TOML file."""
        if is_available("tomllib"):  # pragma: no cover
            import tomllib

            with open(file_path, "rb") as file:
                data = tomllib.load(file)

        elif is_available("tomli"):  # pragma: no cover
            import tomli  # type: ignore

            with open(file_path, "rb") as file:
                data = tomli.load(file)

        elif is_available("tomlkit"):  # pragma: no cover
            import tomlkit  # type: ignore

            with open(file_path) as file:
                data = tomlkit.load(file)

        else:  # pragma: no cover
            raise ImportError(
                "No TOML provider found in the current environment. "
                "Please install tomli using 'pip install tomli'. "
                "Alternatively, run readme-ai using Python 3.11+."
            )
        return {key.lower(): value for key, value in data.items()}

    @staticmethod
    def read_text(file_path: str | Path) -> str:
        """Read the content of a TXT file."""
        with open(file_path, encoding="utf-8") as file:
            return file.read()

    @staticmethod
    def read_yaml(file_path: str | Path) -> dict[str, Any]:
        """Read the content of a YAML file."""
        with open(file_path, encoding="utf-8") as file:
            return yaml.safe_load(file)

    @staticmethod
    def write_html(file_path: str | Path, content: str) -> None:
        """Write the content to an HTML file."""
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(content)

    @staticmethod
    def write_json(file_path: str | Path, content: dict[str, Any]) -> None:
        """Write the content to a JSON file."""
        with open(file_path, "w", encoding="utf-8") as file:
            json.dump(content, file, indent=4)

    @staticmethod
    def write_markdown(file_path: str | Path, content: str) -> None:
        """Write the content to a Markdown file."""
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(content)

    @staticmethod
    def write_toml(file_path: str | Path, content: dict[str, Any]) -> None:
        """Write the content to a TOML file."""
        raise NotImplementedError("Writing to TOML files is not supported.")

    @staticmethod
    def write_text(file_path: str | Path, content: str) -> None:
        """Write the content to a TXT file."""
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(content)

    @staticmethod
    def write_yaml(file_path: str | Path, content: dict[str, Any]) -> None:
        """Write the content to a YAML file."""
        with open(file_path, "w", encoding="utf-8") as file:
            yaml.safe_dump(content, file)



import importlib


def is_available(name: str) -> bool:
    """Check if a module is available for import."""
    try:
        importlib.import_module(name)
        return True
    except ModuleNotFoundError:  # pragma: no cover
        return False



import re


class DocumentCleaner:
    """
    Document cleaner to preprocess repository content.
    """

    def __init__(
        self,
        remove_empty_lines: bool = True,
        remove_extra_whitespaces: bool = True,
        remove_trailing_whitespaces: bool = True,
        normalize_indentation: bool = True,
    ):
        self.remove_empty_lines = remove_empty_lines
        self.remove_extra_whitespaces = remove_extra_whitespaces
        self.remove_trailing_whitespaces = remove_trailing_whitespaces
        self.normalize_indentation = normalize_indentation

    def clean(self, code: str) -> str:
        """Clean the given document string."""
        if self.remove_empty_lines:
            code = self._remove_empty_lines(code)
        if self.remove_extra_whitespaces:
            code = self._remove_extra_whitespaces(code)
        if self.remove_trailing_whitespaces:
            code = self._remove_trailing_whitespaces(code)
        if self.normalize_indentation:
            code = self._normalize_indentation(code)
        return code.strip()

    def _remove_empty_lines(self, code: str) -> str:
        """Remove empty lines and lines with only whitespace."""
        return "\n".join(line for line in code.splitlines() if line.strip())

    def _remove_extra_whitespaces(self, code: str) -> str:
        """Remove extra whitespaces within lines."""
        return re.sub(r"\s+", " ", code)

    def _remove_trailing_whitespaces(self, code: str) -> str:
        """Remove trailing whitespaces from each line."""
        return "\n".join(line.rstrip() for line in code.splitlines())

    def _normalize_indentation(self, code: str) -> str:
        """Normalize indentation to spaces."""
        lines = code.splitlines()
        normalized_lines = []
        for line in lines:
            indent = len(line) - len(line.lstrip())
            normalized_lines.append(" " * indent + line.lstrip())
        return "\n".join(normalized_lines)



import asyncio
import os
import platform
import shutil
from pathlib import Path

import git

from readmeai.errors import GitCloneError
from readmeai.logger import get_logger
from readmeai.preprocessor.directory_cleaner import (
    remove_directory,
    remove_hidden_contents,
)

_logger = get_logger(__name__)


async def clone_repository(
    repo_url: str, target: Path, depth: int = 1
) -> None:
    """Clone a Git repository to the specified target directory."""
    loop = asyncio.get_event_loop()
    await loop.run_in_executor(
        None,
        lambda: git.Repo.clone_from(
            repo_url,
            str(target),
            depth=depth,
            single_branch=True,
        ),
    )


async def copy_directory(source: Path, target: Path) -> None:
    """Copy a directory and its contents to a new location."""
    if platform.system() == "Windows":
        os.system(f'xcopy "{source}" "{target}" /E /I /H /Y')
    else:
        await asyncio.to_thread(
            shutil.copytree,
            source,
            target,
            dirs_exist_ok=True,
            ignore=shutil.ignore_patterns(".git"),
        )


async def load_data(repository: Path | str, temp_dir: str) -> str:
    """Clone repository to temporary directory and return the path."""
    temp_dir_path = Path(temp_dir)
    repo_path = Path(repository)

    try:
        if temp_dir_path.exists():
            await remove_directory(temp_dir_path)

        if repo_path.is_dir():
            await copy_directory(repo_path, temp_dir_path)
        else:
            await clone_repository(str(repository), temp_dir_path)

        await remove_hidden_contents(temp_dir_path)

        return str(temp_dir_path)

    except (
        git.GitCommandError,
        git.InvalidGitRepositoryError,
        git.NoSuchPathError,
    ) as e:
        _logger.error(f"Failed to clone repository {repository}: {e}")
        raise GitCloneError(
            f"Failed to clone repository {repository}",
        ) from e

    except Exception as e:
        _logger.error(
            f"Unexpected error while cloning repository {repository}: {e}",
        )
        raise GitCloneError(
            f"Unexpected error while cloning repository {repository}",
        ) from e



import enum
import functools
from pathlib import Path

from pydantic import (
    BaseModel,
    Field,
    HttpUrl,
    field_validator,
    model_validator,
)

from readmeai.errors import GitURLError


class GitHost(enum.Enum):
    """
    Git repository hosting providers.
    """

    GITHUB = (
        "github.com",
        "https://api.github.com/repos/",
        "https://github.com/{full_name}/blob/main/{file_path}",
    )
    GITLAB = (
        "gitlab.com",
        "https://gitlab.com/api/v4/projects/",
        "https://gitlab.com/{full_name}/-/blob/main/{file_path}",
    )
    BITBUCKET = (
        "bitbucket.org",
        "https://api.bitbucket.org/2.0/repositories/",
        "https://bitbucket.org/{full_name}/src/master/{file_path}",
    )
    LOCAL = ("local", "", "{file_path}")

    def __init__(self, name: str, api_url: str, file_url_template: str):
        self.domain = name
        self.api_url = api_url
        self.file_url_template = file_url_template


class GitURL(BaseModel):
    """
    Git repository URL model with validation and parsing methods.
    """

    url: HttpUrl
    host: GitHost | None = Field(default=None)
    host_domain: str = Field(default="")
    name: str = Field(default="")
    full_name: str = Field(default="")

    model_config = {
        "frozen": True,
        "use_enum_values": True,
        "extra": "forbid",
        "arbitrary_types_allowed": True,
    }

    @classmethod
    @functools.lru_cache(maxsize=100)
    def create(cls, url: HttpUrl) -> "GitURL":
        """Create a GitURL object from a string URL."""
        return cls(url=url)

    @field_validator("url")
    @classmethod
    def validate_url(cls, v: HttpUrl) -> HttpUrl:
        """Validates the Git repository URL."""
        try:
            parse_git_url(str(v))
        except ValueError as e:
            raise ValueError(f"Invalid Git repository URL: {v}") from e
        return v

    @model_validator(mode="after")
    def set_attributes(self) -> "GitURL":
        """Sets the Git URL attributes based on the URL."""
        try:
            host_domain, host, name, full_name = parse_git_url(str(self.url))
            object.__setattr__(self, "host_domain", host_domain)
            object.__setattr__(self, "name", name)
            object.__setattr__(self, "full_name", full_name)
            for git_host in GitHost:
                if git_host.name.lower() == host:
                    object.__setattr__(self, "host", git_host)
                    break
        except ValueError as e:
            raise ValueError(f"Failed to parse Git URL: {self.url}") from e
        return self

    def get_api_url(self) -> str:
        """Return the REST API endpoint URL for a git repository."""
        if self.host is None or self.host == GitHost.LOCAL:
            raise ValueError(
                f"Unsupported Git host or local repository: {self.url}",
            )
        if self.full_name:
            return f"{self.host.api_url}/{self.full_name}"
        else:
            raise ValueError("Repository full name is required.")

    def get_file_url(self, file_path: str) -> str:
        """Return the URL of the file in the remote repository."""
        if self.host:
            return self.host.file_url_template.format(
                full_name=self.full_name,
                file_path=file_path,
            )
        raise ValueError(f"Unsupported Git host: {self.url}")


def parse_git_url(url: str | Path) -> tuple[str, str, str, str]:
    """Parse repository URL and return host, full name, and project name."""
    if isinstance(url, Path) or (isinstance(url, str) and Path(url).is_dir()):
        host_domain = host = GitHost.LOCAL.name
        name = Path(url).name
        full_name = f"{Path(url).parent.name}/{name}"
    else:
        try:
            parsed_url = HttpUrl(url)
            if parsed_url.scheme not in ["http", "https"]:
                raise GitURLError(
                    url,
                    f"Uknown scheme provided: {parsed_url.scheme}",
                )
        except ValueError as e:
            raise GitURLError(url) from e

        assert (
            parsed_url.host and parsed_url.path
        ), f"Invalid Git repository URL: {parsed_url}"

        path_parts = parsed_url.path.strip("/").split("/")

        full_name = "/".join(path_parts[:2])

        host = parsed_url.host.split(".")[0].lower()

        host_domain = parsed_url.host

        name = path_parts[-1]

    return host_domain, host, name, full_name



"""
Parser for gradle dependency files.
"""

import re

from readmeai.parsers.base import BaseFileParser


class BuildGradleParser(BaseFileParser):
    """
    Parser for build.gradle dependency files.
    """

    def __init__(self) -> None:
        """Initializes the handler with given configuration."""
        super().__init__()

    def parse(self, content: str) -> list[str]:
        """Extracts package names from a build.gradle file."""
        try:
            pattern = r"(implementation|classpath|api|testImplementation|androidTestImplementation|kapt)\s+['\"]([^'\"]+)['\"]"
            matches = re.findall(pattern, content)
            package_names = set()
            for _, dependency in matches:
                parts = dependency.split(":")
                for part in parts:
                    part = part.split(".")
                    for p in part:
                        if p.isalpha():
                            package_names.add(p)

            return list(package_names)

        except re.error as exc:
            return self.handle_parsing_error(f"build.gradle: {exc!s}")


class BuildGradleKtsParser(BaseFileParser):
    """
    Parser for build.gradle.kts dependency files.
    """

    def parse(self, content: str) -> list[str]:
        """Extracts package names from a build.gradle.kts file."""
        try:
            pattern = r"(\bimplementation|testImplementation)\s*\((['\"])([^'\"]+)\2\)"
            matches = re.findall(pattern, content)
            package_names = set()
            for _, _, dependency in matches:
                parts = dependency.split(":")
                for part in parts:
                    part = part.split(".")
                    for p in part:
                        if p.isalpha():
                            package_names.add(p)
            return list(package_names)

        except re.error as error:
            return self.handle_parsing_error(f"build.gradle.kts: {error!s}")



"""Image generation using OpenAI's DALL-E model.

Notes:
- Generated image is saved as a PNG file and used as the project logo in the README file.
- Currently, OpenAI is the only supported provider for generating images with readme-ai.
"""

from typing import Any

import aiohttp
import openai

from readmeai.config.constants import ImageOptions
from readmeai.config.settings import ConfigLoader
from readmeai.logger import get_logger


class DalleHandler:
    """
    Generate and download an image using OpenAI's DALL-E model.
    """

    def __init__(self, config: ConfigLoader) -> None:
        self.config = config
        self.default_image = ImageOptions.BLUE.value
        self.filename = f"{config.config.git.name}.png"
        self._logger = get_logger(__name__)
        self._model_settings()

    def _model_settings(self) -> None:
        self.client = openai.AsyncOpenAI()
        self.model = "dall-e-3"
        self.size = "1792x1024"
        self.quality = "standard"
        self.n = 1

    async def __aenter__(self):
        self.session = aiohttp.ClientSession()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.session.close()
        self._logger.debug(f"Closed {self.model.upper()} API session...")

    def _build_payload(self) -> dict[str, Any]:
        """Formats the prompt string using configuration data."""
        return {
            "prompt": self.config.prompts["prompts"]["logo"].format(
                project_name=self.config.config.git.name,
                project_slogan=self.config.config.md.slogan,
                project_overview=self.config.config.md.overview,
            ),
            "model": self.model,
            "size": self.size,
            "quality": self.quality,
            "n": self.n,
        }

    async def _make_request(self) -> str:
        """Generate an image using the DALL-E model, and return its URL."""
        try:
            payload = self._build_payload()
            response = await self.client.images.generate(**payload)
            if response and response.data and response.data[0].url:
                return response.data[0].url
            else:
                self._logger.error(
                    f"Failed to generate project logo image: {response}"
                )
                return self.default_image

        except (Exception, openai.OpenAIError) as e:
            self._logger.error(f"Error generating project logo image: {e!r}")
            return self.default_image

    async def download(self, image_url) -> str:
        """Download the generated image from the provided URL."""
        try:
            async with self.session.get(image_url) as response:
                if response.status == 200:
                    content = await response.read()
                    with open(self.filename, "wb") as f:
                        f.write(content)

                    self._logger.info(
                        f"{self.model.upper()} image successfully downloaded."
                    )
                    return self.filename
                else:
                    self._logger.error(
                        f"Failed to download image: {response.status}"
                    )

        except Exception as e:
            self._logger.error(f"Error downloading image: {e!r}")

        return self.default_image

    async def generate_and_download(self) -> str:
        """Generate and download the image."""
        image_url = await self._make_request()
        return await self.download(image_url)



from abc import ABC, abstractmethod
from typing import Any


class BaseTemplate(ABC):
    """
    Abstract base class for all Markdown templates.
    """

    @abstractmethod
    def render(self, data: dict[str, Any]) -> str: ...



from typing import Any, ClassVar

from readmeai.config.constants import HeaderStyleOptions
from readmeai.templates.base import BaseTemplate


class HeaderTemplate(BaseTemplate):
    """
    Class variable for rendering the README.md header style.
    """

    HEADER_TEMPLATES: ClassVar[dict] = {
        HeaderStyleOptions.ASCII: """\
<div align="{align}">\n{image}\n</div>
<p align="{align}">\n\t<em>{slogan}</em>\n</p>
<p align="{align}">\n\t{shields_icons}</p>
<p align="{align}">{badges_tech_stack_text}</p>
<p align="{align}">\n\t{badges_tech_stack}</p>
<br>
""",
        HeaderStyleOptions.CLASSIC: """\
<p align="{align}">
    <img src="{image}" align="{align}" width="{image_width}">
</p>
<p align="{align}"><h1 align="{align}">{repo_name}</h1></p>
<p align="{align}">\n\t<em>{slogan}</em>\n</p>
<p align="{align}">\n\t{shields_icons}</p>
<p align="{align}">{badges_tech_stack_text}</p>
<p align="{align}">\n\t{badges_tech_stack}</p>
<br>
""",
        HeaderStyleOptions.COMPACT: """\
<div align="left">
    <img src="{image}" width="{image_width}" align="left" style="margin-right: 15px"/>
    <div style="display: inline-block;">
        <h2 style="display: inline-block; vertical-align: middle; margin-top: 0;">{repo_name}</h2>
        <p>\n\t<em>{slogan}</em>\n</p>
        <p>\n\t{shields_icons}</p>
        <p>{badges_tech_stack_text}</p>
        <p>\n\t{badges_tech_stack}</p>
    </div>
</div>
<br clear="left"/>
""",
        HeaderStyleOptions.MODERN: """\
<div align="left" style="position: relative;">
<img src="{image}" align="right" width="{image_width}" style="margin: -20px 0 0 20px;">
<h1>{repo_name}</h1>
<p align="left">\n\t<em>{slogan}</em>\n</p>
<p align="left">\n\t{shields_icons}</p>
<p align="left">{badges_tech_stack_text}</p>
<p align="left">\n\t{badges_tech_stack}</p>
</div>
<br clear="right">
""",
        HeaderStyleOptions.SVG: """\
{image}
<p align="{align}">\n\t<em>{slogan}</em>\n</p>
<p align="{align}">\n\t{shields_icons}</p>
<p align="{align}">{badges_tech_stack_text}</p>
<p align="{align}">\n\t{badges_tech_stack}</p>
<br>
""",
    }

    def __init__(self, style: str = HeaderStyleOptions.CLASSIC) -> None:
        self.style = style

    def render(self, data: dict[str, Any]) -> str:
        """Render the header based on the provided data."""
        template = self.HEADER_TEMPLATES.get(
            self.style,
            self.HEADER_TEMPLATES[HeaderStyleOptions.CLASSIC],
        )
        return template.format(**data)

    @staticmethod
    def get_header_template(template: str) -> str:
        """Get the header template for the given style."""
        return HeaderTemplate.HEADER_TEMPLATES[HeaderStyleOptions(template)]



from pathlib import Path
from string import Template

import tomllib

from readmeai.config.settings import ConfigLoader
from readmeai.generators.quickstart import QuickStartGenerator
from readmeai.ingestion.models import QuickStart, RepositoryContext


class QuickStartBuilder:
    """
    Build the 'Quickstart', or 'Getting Started' README section.
    """

    def __init__(
        self, config_loader: ConfigLoader, repo_context: RepositoryContext
    ):
        self.config_loader = config_loader
        self.config = config_loader.config
        self.repo_context = repo_context
        self.git = self.config.git
        self.md = self.config.md
        self.quickstart_config = self._load_quickstart_config()

    def _load_quickstart_config(self):
        config_path = (
            Path(__file__).parent.parent
            / "config"
            / "settings"
            / "quickstart_config.toml"
        )
        with open(config_path, "rb") as f:
            return tomllib.load(f)

    def build(self) -> str:
        """Create the Installation, Usage, and Testing instructions."""
        repo_url = (
            f"../{self.git.name}"
            if self.git.host_domain.lower() == "local"
            else self.git.repository
        )
        usage_guides = QuickStartGenerator(self.config_loader).generate(
            self.repo_context.language_counts, self.repo_context.metadata
        )

        main_template = Template(self.quickstart_config["templates"]["main"])
        return main_template.safe_substitute(
            prerequisites_section=self._format_prerequisites(usage_guides),
            installation_section=self._format_installation(
                usage_guides, repo_url
            ),
            usage_section=self._format_usage(usage_guides),
            testing_section=self._format_testing(usage_guides),
        )

    def _format_prerequisites(self, usage_guides: QuickStart) -> str:
        template = Template(
            self.quickstart_config["templates"]["prerequisites"]
        )
        return template.safe_substitute(
            repo_name=self.git.name,
            system_requirements=self._format_system_requirements(usage_guides),
        )

    def _format_system_requirements(self, usage_guides: QuickStart) -> str:
        formatting = self.quickstart_config["formatting"]
        requirements = [
            formatting["system_requirements_prefix"].replace(
                "$key", "Programming Language"
            )
            + usage_guides.primary_language
        ]
        if usage_guides.package_managers:
            pkg_managers = ", ".join(
                tool.capitalize() for tool in usage_guides.package_managers
            )
            requirements.append(
                formatting["system_requirements_prefix"].replace(
                    "$key", formatting["package_managers_label"]
                )
                + pkg_managers
            )
        if usage_guides.containers:
            containers = ", ".join(
                tool.capitalize() for tool in usage_guides.containers
            )
            requirements.append(
                formatting["system_requirements_prefix"].replace(
                    "$key", formatting["containers_label"]
                )
                + containers
            )
        return "\n".join(requirements)

    def _format_installation(
        self, usage_guide: QuickStart, repo_url: str
    ) -> str:
        template = Template(
            self.quickstart_config["templates"]["installation"]
        )
        install_steps_template = Template(
            self.quickstart_config["install_steps"]["template"]
        )
        install_steps = install_steps_template.safe_substitute(
            repo_name=self.git.name,
            repo_url=repo_url,
            install_commands=usage_guide.install_commands
            or self.config_loader.tool_config.get("default", {}).get(
                "install", ""
            ),
        )
        return template.safe_substitute(
            repo_name=self.git.name,
            install_steps=install_steps,
        )

    def _format_usage(self, usage_guides: QuickStart) -> str:
        template = Template(self.quickstart_config["templates"]["usage"])
        usage_commands = (
            usage_guides.usage_commands
            or self.config_loader.tool_config.get("default", {}).get(
                "usage", ""
            )
        )
        return template.safe_substitute(
            usage_instructions=f"Run {self.git.name} using the following command:\n{usage_commands.lstrip()}"
        )

    def _format_testing(self, usage_guides: QuickStart) -> str:
        template = Template(self.quickstart_config["templates"]["testing"])
        test_commands = (
            usage_guides.test_commands
            or self.config_loader.tool_config.get("default", {}).get(
                "test", ""
            )
        )
        return template.safe_substitute(
            test_instructions=f"Run the test suite using the following command:\n{test_commands.lstrip()}"
        )



"""Google Gemini LLM API service implementation."""

import os
from typing import Any, Optional

import aiohttp
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from readmeai.config.settings import ConfigLoader
from readmeai.ingestion.models import RepositoryContext
from readmeai.models.base import BaseModelHandler
from readmeai.models.tokens import token_handler

try:
    import google.generativeai as genai

    GENAI_AVAILABLE = True
except ImportError:
    genai = None
    GENAI_AVAILABLE = False


class GeminiHandler(BaseModelHandler):
    """
    Google Gemini LLM API service implementation.
    """

    def __init__(
        self, config_loader: ConfigLoader, context: RepositoryContext
    ) -> None:
        super().__init__(config_loader, context)
        self.model: Optional[Any] = None
        self.model_name: str = "gemini-1.5-flash"
        self.top_p: float = (
            0.8  # Default value, override from config if available
        )
        if GENAI_AVAILABLE:
            self._model_settings()
        else:
            self._logger.warning(
                "Google Generative AI library is not available. Some features will be disabled."
            )

    def _model_settings(self):
        if not GENAI_AVAILABLE:
            self._logger.error(
                "Attempted to configure Gemini model without the required library."
            )
            return

        api_key = os.getenv("GOOGLE_API_KEY")
        if not api_key:
            self._logger.error(
                "GOOGLE_API_KEY environment variable is not set."
            )
            return

        genai.configure(api_key=api_key)
        self.model = genai.GenerativeModel(self.model_name)

        # Safely get top_p from config, use default if not available
        self.top_p = getattr(self.config.llm, "top_p", self.top_p)

    async def _build_payload(self, prompt: str, tokens: int) -> Any:
        """Build payload for POST request to the Gemini API."""
        if not GENAI_AVAILABLE:
            raise RuntimeError(
                "Google Generative AI library is not available."
            )

        return genai.types.GenerationConfig(
            max_output_tokens=tokens,
            temperature=self.temperature,
            top_p=self.top_p,
        )

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_if_exception_type(
            (
                aiohttp.ClientError,
                aiohttp.ClientResponseError,
                aiohttp.ClientConnectorError,
            )
            if GENAI_AVAILABLE
            else tuple()
        ),
    )
    async def _make_request(
        self,
        index: str | None,
        prompt: str | None,
        tokens: int | None,
        repo_files: Any,
    ) -> Any:
        """Processes Gemini API responses and returns generated text."""
        if not GENAI_AVAILABLE:
            self._logger.error(
                "Cannot make request: Google Generative AI library is not available."
            )
            return index, self.placeholder

        if self.model is None:
            self._logger.error("Gemini model is not properly initialized.")
            return index, self.placeholder

        try:
            prompt = await token_handler(self.config, index, prompt, tokens)
            parameters = await self._build_payload(prompt, tokens)

            async with self.rate_limit_semaphore:
                response = await self.model.generate_content_async(
                    prompt,
                    generation_config=parameters,
                )
                data = (
                    response.text
                    if hasattr(response, "text")
                    else str(response)
                )
                self._logger.info(
                    f"Response from Gemini for '{index}': {data}"
                )
                return index, data

        except (
            aiohttp.ClientError,
            aiohttp.ClientResponseError,
            aiohttp.ClientConnectorError,
        ) as e:
            self._logger.error(
                f"Error processing request for '{index}': {e!r}"
            )
            raise  # Re-raise for retry decorator

        except Exception as e:
            self._logger.error(f"Unexpected error for '{index}': {e!r}")
            return index, self.placeholder
